from qtpy import QtCore, QtWidgets

from ScopeFoundry.logged_quantity import LQCollection


class DataBrowserPlugIn(QtCore.QObject):

    """Abstract class for DataBrowserPlugIn"""

    name = "base plugin"  # Override me
    button_text = "☔base"  # override me
    show_keyboard_key = None  # (optional) override me : e.g QtCore.Qt.Key_Q
    description = ""  # (optional) override me, include '(Ctrl + Q)' if applicable

    def setup(self):
        """Override this method"""
        # set self.ui here
        # has self.settings

    def update(self, fname: str = None) -> None:
        """Override this method. This method is called on show plugin and when data_filename has changed."""
        # guard against unsupported files
        # load file, extract information, update ui ...

    def __init__(self, databrowser):
        QtCore.QObject.__init__(self)

        self.databrowser = databrowser
        self.settings = LQCollection()
        self.ui = None
        self.is_loaded = False

        self.show_button = QtWidgets.QPushButton(self.button_text)
        self.show_button.setCheckable(True)
        self.show_button.setChecked(False)
        self.show_button.toggled[bool].connect(self.on_show_toggle)
        self.show_button.setToolTip(f"<b>{self.name}</b> {self.description}")

    @property
    def is_showing(self):
        return self.show_button.isChecked()

    def update_if_showing(self, fname: str):
        if self.is_showing:
            self.update(fname)

    def get_show_hide_button(self):
        btn = self.show_button
        s = f"""QPushButton:!checked{{ background::rgba(216, 223, 235,0.3); border: 1px solid grey; }}
                QPushButton:checked{{ background:rgba(216, 223, 235 ,1); border: 1px solid grey; }}"""
        self.show_button.setStyleSheet(btn.styleSheet() + s)
        return self.show_button

    def toggle_show_hide(self):
        self.show_button.setChecked(not self.show_button.isChecked())

    def show(self):
        if not self.is_loaded:
            self.setup()
            self.is_loaded = True
            if self.ui is not None:
                self.databrowser.ui.data_view_layout.addWidget(self.ui)

        if self.ui is None:
            return

        self.ui.show()

    def hide(self):
        if self.ui is None:
            return
        self.ui.hide()

    def on_show_toggle(self, show):
        if show:
            self.show()
            self.update(self.new_fname)
        else:
            self.hide()

    @property
    def new_fname(self):
        return self.databrowser.settings["data_filename"]



from qtpy import QtCore

from ScopeFoundry import LQCollection


class DataBrowserView(QtCore.QObject):
    """Abstract class for DataBrowser Views"""

    name = "base_view"  # override me! (recommended to use the ScopeFoundry.Measurement.name)

    def __init__(self, databrowser):
        QtCore.QObject.__init__(self)
        self.databrowser = databrowser
        self.settings = LQCollection()
        self.view_loaded = False
        self.ui = None

    def setup(self):
        pass
        # create view with no data file

    def on_change_data_filename(self, fname=None):
        pass
        # load data file

        # update display

    def is_file_supported(self, fname):
        """
        returns whether view can handle file, should return False early to avoid
        too much computation when selecting a file.

        Override me if this fallback does not behave as expected. In particular
            if self.name is not the supported measurement name
            or if supported file is not an h5 file
        """
        return self.check_h5_file_support(fname, (self.name,))

    def check_h5_file_support(self, fname, supported_measurement_names):
        """
        helper function to check if h5 file *fname* has group 'measurement/X'
        where X is any element of *supported_measurement_names*
        """
        if not fname.endswith(".h5"):
            return False

        if isinstance(supported_measurement_names, str):
            supported_measurement_names = (supported_measurement_names,)

        import h5py

        with h5py.File(fname) as file:
            return any(
                name in file["measurement"] for name in supported_measurement_names
            )



from flask import Flask, render_template
from flask_restful import Resource, Api
from qtpy import QtCore
from ScopeFoundry.base_app import BaseMicroscopeApp
from ScopeFoundry.helper_funcs import sibling_path
from collections import OrderedDict

class MicroscopeRestResource(Resource):

    def __init__(self, **kwargs):
        self.microscope_app = kwargs['microscope_app']

    def get(self):
        hardware = OrderedDict()
        for HW in self.microscope_app.hardware.values():
            hardware[HW.name] = dict(settings=OrderedDict([(S.name, S.val) for S in HW.settings.as_list()]))
        measurements = OrderedDict()
        for M in self.microscope_app.measurements.values():
            print(M)
            measurements[M.name] = dict(settings=OrderedDict([(S.name, S.val) for S in M.settings.as_list()]))
        return {'app': dict(name=self.microscope_app.name, settings={S.name: S.val for S in self.microscope_app.settings.as_list()}),
                'hardware': hardware, 
                'measurements': measurements,
                }


def settings_dict(settings):
    out =  OrderedDict([(S.name, S.val) for S in settings.as_list()])
    out['units'] = OrderedDict([(S.name, S.unit) for S in settings.as_list() if S.unit])
    return out

def lq_dict(lq):
    return OrderedDict( [('name', lq.name), ('val',lq.val), ('unit', lq.unit)])


class HardwareSettingsListRestResource(Resource):

    def __init__(self, microscope_app):
        self.microscope_app = microscope_app
        
    def get(self, hw_name):
        settings = self.microscope_app.hardware[hw_name].settings
        return settings_dict(settings)      

class HardwareSettingsLQRestResource(Resource):

    def __init__(self, microscope_app):
        self.microscope_app = microscope_app
        
    def get(self, hw_name, lq_name):
        lq = self.microscope_app.hardware[hw_name].settings.get_lq(lq_name)        
        return lq_dict(lq)

class MeasurementSettingsListRestResource(Resource):

    def __init__(self, microscope_app):
        self.microscope_app = microscope_app
        
    def get(self, measure_name):
        settings = self.microscope_app.measurements[measure_name].settings        
        return settings_dict(settings)
    
class MeasurementSettingsLQRestResource(Resource):

    def __init__(self, microscope_app):
        self.microscope_app = microscope_app
        
    def get(self, measure_name, lq_name):
        lq = self.microscope_app.hardware[measure_name].settings.get_lq(lq_name)        
        return lq_dict(lq)
    
class MicroscopeFlaskWebThread(QtCore.QThread):
    
    def __init__(self, app):
        QtCore.QThread.__init__(self)
        self.app = app
        
        
        self.flask_app = Flask(app.name, template_folder=sibling_path(__file__, 'templates'), )
    
        self.flask_app.route('/')(self.index)
        
        self.rest_api = Api(self.flask_app)
        self.flask_app.config['RESTFUL_JSON'] = dict(indent=4)
        

        self.rest_api.add_resource(MicroscopeRestResource,
                                   '/api/app',
                                   resource_class_kwargs={ 'microscope_app': self.app })
        
        self.rest_api.add_resource(HardwareSettingsListRestResource, 
                                   '/api/hardware/<string:hw_name>/settings', 
                                   resource_class_kwargs={ 'microscope_app': self.app })
        
        self.rest_api.add_resource(HardwareSettingsLQRestResource, 
                                   '/api/hardware/<string:hw_name>/settings/<string:lq_name>', 
                                   resource_class_kwargs={ 'microscope_app': self.app })

        self.rest_api.add_resource(MeasurementSettingsListRestResource, 
                                   '/api/measurements/<string:measure_name>/settings', 
                                   resource_class_kwargs={ 'microscope_app': self.app })
        
        self.rest_api.add_resource(MeasurementSettingsLQRestResource, 
                                   '/api/measurements/<string:measure_name>/settings/<string:lq_name>', 
                                   resource_class_kwargs={ 'microscope_app': self.app })
        for HW in self.app.hardware.values():
            #print(HW.web_ui())
            self.flask_app.route('/hw/{}'.format(HW.name))(HW.web_ui)
        for M in self.app.measurements.values():
            self.flask_app.route('/measure/{}'.format(M.name))(M.web_ui)
    
    def __del__(self):
        self.wait()

    def run(self):
        self.flask_app.run(port=5000)
        print("thread run")

    def index(self):
        return render_template("microscope_index.html", app=self.app)


if __name__ == '__main__':
    import sys
    
    app = BaseMicroscopeApp([])
    app.flask_thread = MicroscopeFlaskWebThread(app)
    app.flask_thread.start()
    sys.exit(app.exec_())


'''
Created on Feb 2, 2023

@author: Benedikt Ursprung
'''
import logging
import sys

from features.import_from_gh_controller import ImportFromGhController
from features.import_from_gh_qt_view import ImportFromGhView
from features.new_hw_controller import NewHWController
from features.new_hw_qt_view import NewHWView
from features.publish_on_gh_controller import PublishOnGhController
from features.publish_on_gh_view import PublishOnGhView
from qtpy.QtWidgets import QApplication, QMainWindow, QTabWidget

logger = logging.getLogger()
logger.addHandler(logging.StreamHandler(sys.stdout))

logger.setLevel(logging.DEBUG)


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.tab_widget = QTabWidget()
        self.setCentralWidget(self.tab_widget)

        self.publish_on_gh_controller = PublishOnGhController(
            PublishOnGhView())
        self.tab_widget.addTab(
            self.publish_on_gh_controller.view, 'publish on GitHub')

        self.import_from_gh_controller = ImportFromGhController(
            ImportFromGhView())
        self.tab_widget.addTab(
            self.import_from_gh_controller.view, 'import from GitHub')

        self.new_hw_controller = NewHWController(NewHWView())
        self.tab_widget.addTab(self.new_hw_controller.view, 'create new HW')

        self.show()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = MainWindow()
    window.setWindowTitle('ScopeFoundry plugin manager')
    sys.exit(app.exec_())



'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''

from ScopeFoundry.base_app import BaseMicroscopeApp


class TestApp(BaseMicroscopeApp):

    name = "$TEST_APP_NAME"

    def setup(self):
        $IMPORTS_IN_TEST_APP
        self.add_hardware($HW_CLASS_NAME(self))
        self.add_measurement($READOUT_CLASS_NAME(self))


if __name__ == '__main__':
    import sys
    app = TestApp(sys.argv)
    sys.exit(app.exec_())



from qtpy.QtWidgets import QLineEdit
from typing_extensions import Self, TypedDict

from .helper_func import new_q_completer
from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem


ITEM_TYPE = "function"


class ExecFunctionKwargs(TypedDict):
    function: str
    args: str


class Function(BaseItem):
    item_type = ITEM_TYPE

    def visit(self) -> None:
        s = "self.app." + self.kwargs["function"] + "(" + self.kwargs["args"] + ")"
        print(eval(s))


class ExecFunctionEditorUI(EditorBaseUI):
    item_type = ITEM_TYPE
    description = "eval a function"

    def __init__(self, measure, all_functions) -> None:
        self.all_functions = all_functions
        super().__init__(measure)

    def setup_ui(self):
        self.function_le = QLineEdit()
        completer = new_q_completer(self.all_functions)
        self.function_le.setCompleter(completer)
        self.function_le.setToolTip("path to a function")
        self.args_le = QLineEdit()
        self.args_le.setToolTip("function arguments")
        self.layout.addWidget(self.function_le)
        self.layout.addWidget(self.args_le)

    def get_kwargs(self) -> ExecFunctionKwargs:
        f = self.function_le.text()
        args = self.args_le.text()
        return {"function": f, "args": args}

    def set_kwargs(self, **kwargs):
        self.function_le.setText(kwargs["function"])
        self.args_le.setText(kwargs["args"])
        self.args_le.selectAll()



import time

from qtpy.QtWidgets import QDoubleSpinBox
from typing_extensions import TypedDict

from .base_item import BaseItem
from .editor_base_ui import EditorBaseUI

ITEM_TYPE = "timeout"


class TimoutKwargs(TypedDict):
    time: float


class Timeout(BaseItem):
    item_type = ITEM_TYPE

    def visit(self):
        t0 = time.time()
        while True:
            dt = time.time() - t0
            if self.measure.interrupt_measurement_called or dt > self.kwargs["time"]:
                break
            time.sleep(0.50)


class TimeoutEditorUI(EditorBaseUI):
    item_type = ITEM_TYPE
    description = "wait for a bit"

    def setup_ui(self):
        time_out_layout = self.group_box.layout()
        self.time_dsb = QDoubleSpinBox()
        self.time_dsb.setValue(0.1)
        self.time_dsb.setToolTip("time-out in sec")
        self.time_dsb.setMaximum(1e6)
        self.time_dsb.setDecimals(3)
        time_out_layout.addWidget(self.time_dsb)

    def get_kwargs(self) -> TimoutKwargs:
        return {"time": self.time_dsb.value()}

    def set_kwargs(self, **kwargs):
        self.time_dsb.setValue(kwargs["time"])
        self.time_dsb.selectAll()
        self.time_dsb.setFocus()



from qtpy import QtWidgets
import h5py

from ScopeFoundry.data_browser import DataBrowserView

class H5TreeView(DataBrowserView):

    name = 'h5_tree'
    
    def is_file_supported(self, fname):
        return ('.h5' in fname)
    
    def setup(self):
        self.ui = QtWidgets.QTextEdit("file_info")
    
    def on_change_data_filename(self, fname=None):
        self.ui.setText("loading {}".format(fname))
        try:
            self.tree_str = "{}\n{}\n".format(fname, "="*len(fname)) 
            with h5py.File(self.fname, 'r') as file:
                file.visititems(self._visitfunc)
            self.ui.setText(self.tree_str)
            
        except Exception as err:
            self.databrowser.ui.statusbar.showMessage("failed to load %s:\n%s" %(fname, err))
            raise(err)

    def _visitfunc(self, name, node):
        
        level = len(name.split('/'))
        indent = '    '*level
        localname = name.split('/')[-1]
    
        if isinstance(node, h5py.Group):
            self.tree_str += indent +"|> {}\n".format(localname)
        elif isinstance(node, h5py.Dataset):
            self.tree_str += indent +"|D {}: {} {}\n".format(localname, node.shape, node.dtype)
        for key, val in node.attrs.items():
            self.tree_str += indent+"    |- {} = {}\n".format(key, val)  
            
            
class H5TreeSearchView(DataBrowserView):
    
    name = 'h5_tree_search'
    
    def is_file_supported(self, fname):
        return ('.h5' in fname)
    
    def setup(self):
        
        #self.settings.New('search_text', dtype=str, initial="")
        
        self.ui = QtWidgets.QWidget()
        self.ui.setLayout(QtWidgets.QVBoxLayout())
        self.search_lineEdit = QtWidgets.QLineEdit()
        self.tree_textEdit = QtWidgets.QTextEdit("")
        
        self.ui.layout().addWidget(self.search_lineEdit)
        self.ui.layout().addWidget(self.tree_textEdit)
         
        #self.settings.search_text.connect_to_widget(self.search_lineEdit)
        #self.settings.search_text.add_listener(self.on_new_search_text)
        self.search_text = ""
        
        self.search_lineEdit.textChanged.connect(self.on_new_search_text)
        

    def on_change_data_filename(self, fname=None):
        self.tree_textEdit.setText("loading {}".format(fname))
        try:
            self.fname = fname        
            self.on_new_search_text()
            self.databrowser.ui.statusbar.showMessage("")
            
        except Exception as err:
            msg = "Failed to load %s:\n%s" %(fname, err)
            self.databrowser.ui.statusbar.showMessage(msg)
            self.tree_textEdit.setText(msg)
            raise(err)


    def on_new_search_text(self, x=None):
        if x is not None:
            self.search_text = x.lower()
        old_scroll_pos = self.tree_textEdit.verticalScrollBar().value()
        self.tree_str = ""
        
        with h5py.File(self.fname, 'r') as file:
            file.visititems(self._visitfunc)
        
        self.tree_text_html = \
        """<html><b>{}</b><hr/>
        <div style="font-family: Courier;">
         {} 
         </div>
         </html>""".format(self.fname, self.tree_str)
        
        self.tree_textEdit.setText(self.tree_text_html)
        self.tree_textEdit.verticalScrollBar().setValue(old_scroll_pos)
           
            
    def _visitfunc(self, name, node):
        
        level = len(name.split('/'))
        indent = '&nbsp;'*4*(level-1)
    
        #indent = '<span style="color:blue;">'.format(level*4)
        localname = name.split('/')[-1]
        
        #search_text = self.settings['search_text'].lower()
        search_text = self.search_text
        if search_text and (search_text in localname.lower()):
            localname = """<span style="color: red;">{}</span>""".format(localname)
    
        if isinstance(node, h5py.Group):
            self.tree_str += indent +"|> <b>{}/</b><br/>".format(localname)
        elif isinstance(node, h5py.Dataset):
            self.tree_str += indent +"|D <b>{}</b>: {} {}<br/>".format(localname, node.shape, node.dtype)
        for key, val in node.attrs.items():
            if search_text:
                if search_text in str(key).lower(): 
                    key = """<span style="color: red;">{}</span>""".format(key)
                if search_text in str(val).lower(): 
                    val = """<span style="color: red;">{}</span>""".format(val)
            self.tree_str += indent+"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;|- <i>{}</i> = {}<br/>".format(key, val)          
    



import sys
import time
import logging
import numpy as np

from ScopeFoundry import BaseMicroscopeApp  
from ScopeFoundry.examples.hardware.dummy_xy_stage import DummyXYStageHW
from ScopeFoundry.examples.hardware.dummy_xyz_stage import DummyXYZStageHW
from ScopeFoundry.scanning import BaseRaster2DSlowScan, BaseRaster2DFrameSlowScan, BaseRaster3DSlowScan

logging.basicConfig(level=logging.INFO)  # , filename='m3_log.txt')
logging.getLogger('ScopeFoundry').setLevel(logging.DEBUG)


class TestRaster2DSlowScan(BaseRaster2DSlowScan):
    name = 'test_cart_2d_slow_scan'
    
    def __init__(self, app):
        BaseRaster2DSlowScan.__init__(self, app, h_limits=(0, 100), v_limits=(0, 100), h_unit="um", v_unit="um")        
    
    def setup(self):
        BaseRaster2DSlowScan.setup(self)
        # self.settings.New('pixel_time', initial=0.001, unit='s', si=False, spinbox_decimals=5)
        self.settings.pixel_time.change_readonly(False) 
    
    def pre_scan_setup(self):
        self.display_update_period = 0.050  # seconds
        
        self.stage = self.app.hardware['dummy_xy_stage']
        if self.settings['save_h5']:
            self.test_data = self.h5_meas_group.create_dataset('test_data', self.scan_shape, dtype=float)
        
        self.prev_px = time.time()
         
    def post_scan_cleanup(self):
        print("post_scan_cleanup")
        
    def collect_pixel(self, pixel_i, k, j, i):
        # print pixel_i, k,j,i
        t0 = time.time()
        # px_data = np.random.rand()
        # px_data = t0 - self.prev_px
        x0, y0 = self.pos
        x_set = self.stage.settings['x_position']
        y_set = self.stage.settings['y_position']
        x_hw = self.stage.settings.x_position.read_from_hardware(send_signal=False)
        y_hw = self.stage.settings.y_position.read_from_hardware(send_signal=False)
        if np.abs(x_hw - x0) > 1:
            self.log.debug('=' * 60)
            self.log.debug('pos      {} {}'.format(x0, y0))
            self.log.debug('settings {} {}'.format(x_set, y_set))
            self.log.debug('hw       {} {}'.format(x_hw, y_hw))            
            self.log.debug('settings value delta {} {}'.format(x_set - x0, y_set - y0))
            self.log.debug('read_hw  value delta {} {}'.format(x_hw - x0, y_hw - y0))
            self.log.debug('=' * 60)
        
        x = x_hw
        y = y_hw
        
        px_data = np.sinc((x - 50) * 0.05) ** 2 * np.sinc(0.05 * (y - 50)) ** 2  # + 0.05*np.random.random()
        # px_data = (x-xhw)**2 + ( y-yhw)**2
        # if px_data > 1:
        #    print('hw', x, xhw, y, yhw)
        self.display_image_map[k, j, i] = px_data
        if self.settings['save_h5']:
            self.test_data[k, j, i] = px_data 
        time.sleep(self.settings['pixel_time'])
        # self.prev_px = t0


class TestRaster2DFrameSlowScan(BaseRaster2DFrameSlowScan):
    name = 'test_cart_2d_frame_slow_scan'
    
    def __init__(self, app):
        BaseRaster2DSlowScan.__init__(self, app, h_limits=(0, 100), v_limits=(0, 100), h_unit="um", v_unit="um")        
        
    def scan_specific_setup(self):  # called on app initialization
        self.settings.pixel_time.change_readonly(False)
    
    def pre_scan_setup(self):  # called at the begining of run 
        self.display_update_period = 0.050  # seconds
        
        self.stage = self.app.hardware['dummy_xy_stage']
        if self.settings['save_h5']:
            self.test_data = self.create_h5_framed_dataset("test_data", self.display_image_map)
        
        self.prev_px = time.time()
         
    def post_scan_cleanup(self):  # called at the end of run 
        print("post_scan_cleanup")
        
    def collect_pixel(self, pixel_i, frame_i, k, j, i):
        t0 = time.time()
        x_hw = self.stage.settings.x_position.read_from_hardware(send_signal=False)
        y_hw = self.stage.settings.y_position.read_from_hardware(send_signal=False)
        
        theta = np.pi / 10. * frame_i
        x = x_hw * np.cos(theta) - y_hw * np.sin(theta)
        y = x_hw * np.sin(theta) + y_hw * np.cos(theta)
        
        px_data = np.sinc((x - 50) * 0.05) ** 2 * np.sinc(0.05 * (y - 50)) ** 2  # + 0.05*np.random.random()

        self.display_image_map[k, j, i] = px_data
        if self.settings['save_h5']:
            self.test_data[frame_i, k, j, i] = px_data 
        time.sleep(self.settings['pixel_time'])
        
    def on_new_frame(self, frame_i):
        print("on_new_frame")
        if self.settings['save_h5']:
            self.extend_h5_framed_dataset(self.test_data, frame_i)
            
    def on_end_frame(self, frame_i):
        pass


class TestRaster3DSlowScan(BaseRaster3DSlowScan):
    name = 'test_cart_3d_slow_scan'
    
    def __init__(self, app):
        BaseRaster3DSlowScan.__init__(self, app, h_limits=(0, 100), v_limits=(0, 100), z_limits=(-1, 1), h_unit="um", v_unit="um", z_unit="um")        
    
    def setup(self):
        BaseRaster3DSlowScan.setup(self)
        # self.settings.New('pixel_time', initial=0.001, unit='s', si=False, spinbox_decimals=5)
        self.settings.pixel_time.change_readonly(False) 
    
    def pre_scan_setup(self):
        self.display_update_period = 0.050  # seconds
        
        self.stage = self.app.hardware['dummy_xyz_stage']
        if self.settings['save_h5']:
            self.test_data = self.h5_meas_group.create_dataset('test_data', self.scan_shape, dtype=float)
        
        self.prev_px = time.time()
         
    def post_scan_cleanup(self):
        print("post_scan_cleanup")
        
    def collect_pixel(self, pixel_i, k, j, i):
        
        t0 = time.time()
        
        x0, y0 = self.pos
        x_set = self.stage.settings['x_position']
        y_set = self.stage.settings['y_position']

        x_hw = self.stage.settings.x_position.read_from_hardware(send_signal=False)
        y_hw = self.stage.settings.y_position.read_from_hardware(send_signal=False)

        if np.abs(x_hw - x0) > 1:
            self.log.debug('=' * 60)
            self.log.debug('pos      {} {}'.format(x0, y0))
            self.log.debug('settings {} {}'.format(x_set, y_set))
            self.log.debug('hw       {} {}'.format(x_hw, y_hw))            
            self.log.debug('settings value delta {} {}'.format(x_set - x0, y_set - y0))
            self.log.debug('read_hw  value delta {} {}'.format(x_hw - x0, y_hw - y0))
            self.log.debug('=' * 60)
        
        x = x_hw
        y = y_hw
        
        px_data = np.sinc((x - 50) * 0.05) ** 2 * np.sinc(0.05 * (y - 50)) ** 2

        # NOTE: In 3D index k represents z position / while in 2D k represents 'frame'
        self.display_image_map[k, j, i] = px_data
        if self.settings['save_h5']:
            self.test_data[k, j, i] = px_data 
        time.sleep(self.settings['pixel_time'])
        

class TestRaster2DSlowScanApp(BaseMicroscopeApp):
    
    name = 'app'

    def setup(self):
        
        stage = self.add_hardware(DummyXYStageHW(self))
        stage.settings['connected'] = True
        self.add_measurement(TestRaster2DSlowScan(self))
        self.add_measurement(TestRaster2DFrameSlowScan(self))

        stage = self.add_hardware(DummyXYZStageHW(self))
        self.add_measurement(TestRaster3DSlowScan(self))
        
        for m in self.measurements.values():
            S = m.settings
            # S = self.test_cart_2d_slow_scan.settings
            S['h1'] = 100
            S['v1'] = 100
            S['dh'] = S['dv'] = 1.0
        

if __name__ == '__main__': 
    app = TestRaster2DSlowScanApp([])
    
    # from ScopeFoundry.flask_web_view.flask_web_view import MicroscopeFlaskWebThread
    # app.flask_thread = MicroscopeFlaskWebThread(app)
    # app.flask_thread.start()
    
    sys.exit(app.exec_())    





import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtCore
import numpy as np
import math
import threading

# TODO 
#  connect color scales together DONE (If connected to a HistLut)
#  always show lowest zoom behind DONE
# show one lower zoom underneath


class TileItem(object):
    
    def __init__(self, zmi, zoom, ii, jj):
        self.zmi = zmi
        self.image_item = None
        self.zoom_tile_coord = zoom, ii, jj
        
        self.update_data()
        
        self.transform = QtGui.QTransform()
        
        self.rebuild_data = False
        self.visible = False
    
    def update_data(self):
        zoom, ii,jj = self.zoom_tile_coord
        # zoom
        # zoom z=0 --> original size: tile slice (ii*tile_size : (ii+1)*tile_size : 1)
        # zoom z=1 --> half size: tile slice (ii*tile_size*2 : (ii+1)*tile_size*2 : 2)
        # zoom z=N ---> tile slice (ii*tile_size*(2**z) : (ii+1)*tile_size*(2**z) : (2**z) )
        z = zoom
        zf = 2**z
        Nt = self.zmi.tile_size

        tile_shape = list(self.zmi.image.shape)
        tile_shape[0] = Nt
        tile_shape[1] = Nt
        self.data = np.zeros(tuple(tile_shape), dtype=self.zmi.image.dtype)
        if self.zmi.fill is not None:
            self.data += self.zmi.fill
        
        #print("---", ii, jj, Nt, zoom, zf)
        im = self.zmi.image[ii*Nt*zf:(ii+1)*Nt*zf:zf,
                            jj*Nt*zf:(jj+1)*Nt*zf:zf]
        self.data[:im.shape[0], :im.shape[1]] = im

        
    def set_visible(self, vis=True, force_replacement=False):
        
        #print("set_visible", self.zoom_tile_coord, vis)

        if (not vis) or (not self.image_item is None) or force_replacement:
            self.zmi.plot.removeItem(self.image_item)
            self.image_item = None
            self.visible = False
        
        if vis and (self.image_item is None):
            self.image_item = pg.ImageItem(self.data, **self.zmi.im_kwargs)
            #self.image_item.setRect(pg.QtCore.QRectF(*self.rect))
            self.image_item.setTransform(self.recomputeTransform())
            self.image_item.setZValue(self.zmi.z_value-self.zoom_tile_coord[0])
            self.zmi.plot.addItem(self.image_item)
            self.visible = True
    
    
    def recomputeTransform(self):
        zoom, ii, jj = self.zoom_tile_coord
        zf = 2**zoom
        
        Nx, Ny = self.zmi.image.shape[0:2]
        
        Nt = self.zmi.tile_size 
        
        x = ii*zf* Nt
        y = jj*zf* Nt
    
        tr = self.tile_transform = QtGui.QTransform()
        
        sx = zf#*Nt/Nx * Nt
        sy = zf#*Nt/Ny * Nt
        tr.translate(x, y)
        tr.scale(sx, sy)
        
        #print(f"Nt={Nt}, z={z}, zf={zf}, ii={ii}, jj={jj}, x={x}, y={y}, scale {sx} {sy}")
    
        # full transform 
        self.full_transform = self.tile_transform * self.zmi.transform
        return self.full_transform

        
    
class ZoomableMapImageItem(QtCore.QObject):
    
    sigImageChanged = QtCore.Signal()

    
    def __init__(self, plot_item, image=None, rect=None, transform=None, 
                 tile_size=256, z_value=100, fill=None,
                 **kwargs):
        
        QtCore.QObject.__init__(self)
        
        """
        A Zoomable Map (image pyramid) that acts like a pyqtgraph ImageItem
        for high performance viewing of gigapixel-sized datasets
        
        plot_item: pyqtgraph PlotItem where Zoomable Map will appear
        
        image: large image array that will be shown        
        
        rect: Bounding rectangle for image
            if rect=None, go by pixels otherwise rect=(x0,y0,w,h)
    
        transform: a QTransform that will override rect and allows for 
            arbitrary matrix transforms of the coordinate system
        
        z_value: z-stacking value of highest resolution layer, 
                    different zoom levels are at z_value - zoom
        
        fill: value to place on edge tiles that extend beyond original image
        
        kwargs are sent to setImage of imageItems
        """
        
        self.tile_size = tile_size
        self.z_value = z_value
        self.fill = fill
        self.im_kwargs = kwargs

        self.plot = plot_item
        self.vb  = self.plot.getViewBox()
        
        self.tile_cache = dict()
        self.visible_tiles = dict()
        
        self.lut = kwargs.get('lut', None)
        
        
        #self.plot.sigRangeChanged.connect(self.on_range_changed)
        # Use a signalproxy to limit calls to recompute tiles
        self.sigprox = pg.SignalProxy(
                            signal=self.plot.sigRangeChanged,
                            #delay=0.1,
                            rateLimit=30,
                            slot=self.on_range_changed)

        
        self.image = image
        
        if transform is None:
            self.transform = QtGui.QTransform()
            if not (rect is None):
                self.setRect(rect)
        else: 
            self.transform = transform

        self.lock = threading.Lock()
        
        z_max = self.get_max_zoom()
        #print('z_max', z_max)
        self.get_tile(z_max, 0,0).set_visible()

    def clear_tile_cache(self, clear_visible=False):
        pass
        #TODO
                   
    def get_tile(self, z, ii,jj):
        if (z, ii,jj) not in self.tile_cache:
            self.tile_cache[(z,ii,jj)] = TileItem(self, z, ii, jj)
        return self.tile_cache[(z,ii,jj)]

        
    def tiles_at_zoom(self, z):
        zfi = 2**z
        Nt = self.tile_size
        Nx_tiles = int(np.ceil(self.image.shape[0]/zfi/Nt))
        Ny_tiles = int(np.ceil(self.image.shape[1]/zfi/Nt))
        return (Nx_tiles, Ny_tiles)
    

    def get_max_zoom(self):
        
        # FIXME Need to check this
        
        # Find the zoom where Nx_tiles or Ny_tiles goes to 1
        # Nx_tiles = (im.shape[0]/zfi/Nt)
        zf_max = np.max(self.image.shape[0:1])/self.tile_size 
        z_max = int(math.ceil(math.log2(zf_max)))
        
        z_max = max(0, z_max)
        return z_max
    
    def on_range_changed(self, src, new_range=None):
        if new_range is None:
            new_range = self.vb.viewRange()
        
        tr = self.transform
        tI,invertible = tr.inverted()
        

        # Use the vector of the bottom-left to top-right of view as a measuring stick
        # in all coordinate systems to determine zoom factor and visibility of tiles
        
        (xl,xr), (yb, yt) = new_range  # coordinates in View Space
        
        # pixel size in view coordinate system
        px, py = self.vb.viewPixelSize()
        
        # Coordinates in Image original pixel space
        x0, y0 = tI.map(xl,yb)
        x1, y1 = tI.map(xr,yt)
        
        
        #
        bl_tr_dist_view = np.sqrt( (xr-xl)**2 + (yt-yb)**2)
        bl_tr_dist_view_px = np.sqrt( ((xr-xl)/px)**2 + ((yt-yb)/py)**2)
        bl_tr_dist_orig = np.sqrt( (x1-x0)**2 + (y1-y0)**2)
        
        # print("bl_tr_dist_view", bl_tr_dist_view)
        # print("bl_tr_dist_view_px", bl_tr_dist_view_px)
        # print("bl_tr_dist_orig", bl_tr_dist_orig)

        zf = bl_tr_dist_orig/bl_tr_dist_view_px
        

        z_max = self.get_max_zoom()
        # print("zoom factor", zf, 'z_max', z_max)
        
        zoom = min(max(int(np.floor(np.log2(zf))),0), z_max)
        zfi = 2**zoom # integer power of two zoom factor    

        # Num of tiles possible at current zoom
        Nx_tiles, Ny_tiles = self.tiles_at_zoom(zoom)
        Nt = self.tile_size        
        
        #print("zoom", zoom,'zfi', zfi, "Tiles at zoom", Nx_tiles, Ny_tiles )

        # Tile limits ii0 to ii1, jj0 to jj1
        ii0 = x0 / (zfi*Nt)
        ii0 = max(0, math.floor(ii0)) # clip at zero
        ii1 = x1 / (zfi*Nt)
        ii1 = min(math.ceil(ii1), Nx_tiles)
    
        jj0 = y0 / (zfi*Nt)
        jj0 = max(0, math.floor(jj0)) # clip at zero
        jj1 = y1 / (zfi*Nt)
        jj1 = min(math.ceil(jj1), Ny_tiles)    

                
        #print(f"allowed tiles zoom level {zoom}, x: {ii0} to {ii1}, y: {jj0} to {jj1}")
        
        if self.lock.locked():
            return
        with self.lock:
            # find tiles in visible_tiles that are not allowed and set_visible False
            for (zt, iit, jjt) in list(self.visible_tiles.keys()):
                tile = self.visible_tiles[(zt, iit, jjt)]
                vis = True
                if zt != zoom: # Wrong zoom
                    vis = False
                else:
                    if not ii0 <= iit < ii1: # out of range x
                        vis = False
                    if not jj0 <= jjt < jj1: # out of range y
                        vis = False            
                if not vis:
                    tile.set_visible(False)
                    del self.visible_tiles[(zt, iit, jjt)]
            
            # find or create allowed Tiles in tile cache
            for ii in range(ii0,ii1):
                for jj in range(jj0,jj1):
                    t = self.get_tile(zoom, ii, jj)
                    # set tiles to visible (add image item)
                    t.set_visible(True)
                    # add tiles to visible_tiles dictionary
                    self.visible_tiles[(zoom,ii,jj)] = t
        
        Nx, Ny = self.tiles_at_zoom(z_max)
        for ii in range(Nx):
            for jj in range(Ny):
                self.get_tile(z_max, ii,jj).set_visible(True)
        
        N_visible = len(self.visible_tiles)
        #print(f"zf={zf} z={zoom}, zfi={zfi}. Tiles {Nx_tiles} x {Ny_tiles}, {N_visible} tiles shown")


    # The following methods act like the equivalent single ImageItem methods 

    def getHistogram(self, bins='auto', step='auto', perChannel=False, targetImageSize=200,
                     targetHistogramSize=500, **kwds):
        z_max = self.get_max_zoom()
        # Use biggest tile to get histogram
        return self.get_tile(z_max,0,0).image_item.getHistogram()
            
    def setLookupTable(self, lut, update=True):
        if lut is not self.lut:
            self.lut = lut
            self.im_kwargs['lut'] = lut
        for (zt, iit, jjt) in list(self.visible_tiles.keys()):
            tile = self.visible_tiles[(zt, iit, jjt)]
            tile.image_item.setLookupTable(lut,update)

    def setLevels(self, levels, update=True):
        # if not 
        self.im_kwargs['levels'] = levels
        for (zt, iit, jjt) in list(self.visible_tiles.keys()):
            tile = self.visible_tiles[(zt, iit, jjt)]
            tile.image_item.setLevels(levels,update)
            
    def setRect(self, rect):
        """Scale and translate the image to fit within rect (must be a QRect or QRectF) or iterable [left top width hight]."""
        if not isinstance(rect, QtCore.QRect) or  isinstance(rect, QtCore.QRect):
            rect = QtCore.QRectF(*rect)
        tr = QtGui.QTransform()
        tr.translate(rect.left(), rect.top())
        tr.scale(rect.width() / self.image.shape[0], rect.height() / self.image.shape[1])
        self.setTransform(tr)

    def setTransform(self, tr):
        self.transform = tr
        for (zt, iit, jjt) in list(self.visible_tiles.keys()):
            tile = self.visible_tiles[(zt, iit, jjt)]
            tile.recomputeTransform()

    def setImage(self, image=None, autoLevels=None, slice_changed=None, **kargs):

        self.image = image
        
        for tile_coord in list(self.tile_cache.keys()):
            tile = self.tile_cache[tile_coord]
            if tile.visible:
                tile.update_data()
                tile.set_visible(vis=True, force_replacement=True)
            else:
                del self.tile_cache[tile_coord]
        
        self.sigImageChanged.emit()

        # TODO setImage use slice_changed to limit updates 

    @property
    def levels(self):
        return self.getLevels()
    
    def getLevels(self):
        z_max = self.get_max_zoom()
        return self.get_tile(z_max,0,0).image_item.getLevels()
        




from .base_raster_scan import BaseRaster2DScan, BaseRaster3DScan
from .base_raster_slow_scan import BaseRaster2DSlowScan, BaseRaster3DSlowScan
from .base_raster_frame_slow_scan import BaseRaster2DFrameSlowScan


import random
from ScopeFoundry import HardwareComponent


class DummmyXYZStageDevice:
    
    def __init__(self, debug=False):
        self.x = 0
        self.y = 0
        self.z = 0
        self.debug = debug
        # communicate with hardware here
    
    def read_x(self):
        self.x = self.x + self.noise()
        return self.x

    def read_y(self):
        self.y = self.y + self.noise()
        return self.y

    def read_z(self):
        self.z = self.z + self.noise()
        return self.z
    
    def write_x(self, x):
        self.x = x

    def write_y(self, y):
        self.y = y

    def write_z(self, z):
        self.z = z
    
    def close(self):
        pass
    
    def noise(self):
        return (random.random() - 0.5) * 10e-3


class DummyXYZStageHW(HardwareComponent):
    
    name = "dummy_xyz_stage"
    
    def setup(self):
        lq_params = dict(dtype=float, ro=False,
                           initial=-1,
                           vmin=-1,
                           vmax=100,
                           si=False,
                           unit='um')
        
        self.x_position = self.settings.New("x_position", **lq_params)
        self.y_position = self.settings.New("y_position", **lq_params)       
        self.z_position = self.settings.New("z_position", **lq_params)       

        self.x_position.reread_from_hardware_after_write = True
        self.x_position.spinbox_decimals = 3
        
        self.y_position.reread_from_hardware_after_write = True
        self.y_position.spinbox_decimals = 3

        self.z_position.reread_from_hardware_after_write = True
        self.z_position.spinbox_decimals = 3

    def connect(self):

        dev = self.stage_device = DummmyXYZStageDevice(debug=self.debug_mode.val)

        self.x_position.connect_to_hardware(dev.read_x, dev.write_x)
        self.y_position.connect_to_hardware(dev.read_y, dev.write_y)
        self.z_position.connect_to_hardware(dev.read_z, dev.write_z)

    def disconnect(self):
        
        self.settings.disconnect_all_from_hardware()

        if hasattr(self, 'stage_device'):
            self.stage_device.close()           
            del self.stage_device




import numpy as np
from qtpy.QtWidgets import QComboBox, QDoubleSpinBox
from typing_extensions import TypedDict, Protocol

from .editor_base_controller import EditorBaseController
from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem, VisitReturnType
from .helper_func import new_q_completer


class StartIterationDict(TypedDict):
    values: list
    setting: str


class StartIterationItem(BaseItem):
    item_type = "start-iteration"

    def __init__(self, measure, **kwargs):
        self.iter_id = kwargs["iter_id"]
        self.values = kwargs["values"]
        self._name = kwargs["setting"]
        self.lq = measure.app.get_lq(kwargs["setting"])
        BaseItem.__init__(self, measure=measure, **kwargs)
        self.reset()
        self.end_item = None
        self.val = 0

    def _update_appearance(self, text=None):
        # text = BaseItem._update_appearance(self, text=text)
        t = f"START {self.iter_id}| iter {self._name} over {self.values} ({self.val})"
        self.setText(t)
        return t

    def visit(self) -> VisitReturnType:
        self.idx += 1
        if self.idx == len(self.values) - 1:
            # next time end-iteration is visited the loop breaks
            self.end_item.break_next = True
        self.lq.update_value(self.values[self.idx])
        self.update_text()
        self.measure.iter_values.update({self.iter_id: self.values[self.idx]})
        self.val = self.values[self.idx]

    def reset(self):
        self.idx = -1
        self.update_text()

    def set_end_iteration_item(self, end_item):
        self.end_item = end_item

    def update_text(self):
        text = self.text().split(" - ")[0]
        pct = 100.0 * (self.idx + 1) / (len(self.values))
        if self.idx >= 0:
            texts = [text, f"({self.values[self.idx]})", f"{pct: 1.0f}%"]
        else:
            texts = [text, f"{pct: 1.0f}%"]
        self.setText(" - ".join(texts))


class EndIterationItem(BaseItem):
    item_type = "end-iteration"

    def __init__(self, measure, **kwargs):
        self.kwargs = kwargs
        self.iter_id = kwargs["iter_id"]
        self.values = kwargs["values"]
        self._name = kwargs["setting"]
        self.lq = measure.app.get_lq(kwargs["setting"])
        BaseItem.__init__(self, measure=measure, **kwargs)
        self.reset()
        self.start_item = None
        self.val = 0

    def _update_appearance(self, text=None):
        text = f"END {self.iter_id}| iter {self._name} over {self.values} ({self.val})"
        self.setText(text)
        return text

    def visit(self) -> VisitReturnType:
        self.update_text()
        if self.break_next:
            self.start_item.reset()
            self.reset()
            return None
        else:
            return self.start_item

    def reset(self):
        self.break_next = False
        self.update_text()

    def set_start_iteration_item(self, start_item: StartIterationItem):
        self.start_item = start_item
        self.iter_id = start_item.iter_id
        self._update_appearance()

    def update_text(self):
        try:
            text = self.text().split(" - ")[0]
            item = self.start_item
            pct = item.idx + 1 / len(item.values) * 100
            self.setText(f"{text} - {pct: 1.0f}%")
        except AttributeError:
            pass


class IterationsEditorUI(EditorBaseUI):
    item_type = "start-iteration"
    description = "a setting is iterated over a range of values"

    def __init__(self, measure, paths) -> None:
        self.paths = paths
        super().__init__(measure)

    def setup_ui(self):
        self.setting_cb = QComboBox()
        self.setting_cb.setEditable(True)
        self.setting_cb.addItems(self.paths)
        self.setting_cb.setToolTip("setting")
        completer = new_q_completer(self.paths)
        self.setting_cb.setCompleter(completer)
        self.layout.addWidget(self.setting_cb)
        self.start_dsb = QDoubleSpinBox()
        self.start_dsb.setToolTip("start value")
        self.layout.addWidget(self.start_dsb)
        self.stop_dsb = QDoubleSpinBox()
        self.stop_dsb.setToolTip("stop value")
        self.stop_dsb.setValue(10)
        self.layout.addWidget(self.stop_dsb)
        self.step_dsb = QDoubleSpinBox()
        self.step_dsb.setToolTip("step value")
        self.step_dsb.setValue(1)
        self.layout.addWidget(self.step_dsb)
        for sp in [self.start_dsb, self.step_dsb, self.stop_dsb]:
            sp.setMinimum(-1e6)
            sp.setMaximum(1e6)
            sp.setDecimals(6)

    def get_kwargs(self) -> StartIterationDict:
        path = self.setting_cb.currentText()
        start = self.start_dsb.value()
        stop = self.stop_dsb.value()
        step = self.step_dsb.value()
        values = list(np.arange(start, stop, step))
        return {"setting": path, "values": values}

    def set_kwargs(self, **kwargs):
        self.start_dsb.setValue(kwargs["values"][0])
        step = kwargs["values"][1] - kwargs["values"][0]
        self.step_dsb.setValue(step)
        self.stop_dsb.setValue(kwargs["values"][-1] + step)
        self.setting_cb.setCurrentText(kwargs["setting"])
        self.start_dsb.selectAll()
        self.start_dsb.setFocus()


class InterationsEditorController(EditorBaseController):
    def on_new_func(self):
        kwargs = self.ui.get_kwargs()
        kwargs["iter_id"] = self.measure.next_iter_id()
        self.item_list.add(StartIterationItem(self.measure, **kwargs))
        self.item_list.add(EndIterationItem(self.measure, **kwargs))
        link_iteration_items(self.item_list)

    def on_replace_func(self):
        cur_item = self.item_list.get_current_item()

        if isinstance(cur_item, StartIterationItem):
            start_item, end_item = cur_item, cur_item.end_item
        elif isinstance(cur_item, EndIterationItem):
            start_item, end_item = cur_item.start_item, cur_item
        else:
            return

        kwargs = self.ui.get_kwargs()
        kwargs["iter_id"] = cur_item.iter_id
        self.item_list.replace(StartIterationItem(self.measure, **kwargs), start_item)
        self.item_list.replace(EndIterationItem(self.measure, **kwargs), end_item)
        link_iteration_items(self.item_list)


def link_iteration_items(item_list) -> bool:
    """returns if the list is valid in terms iteration items"""
    start_iter_items: list[StartIterationItem] = []
    for i in range(item_list.count()):
        item = item_list.get_item(i)
        if isinstance(item, StartIterationItem):
            start_iter_items.append(item)
        if isinstance(item, EndIterationItem):
            s_item = start_iter_items.pop()
            item.set_start_iteration_item(s_item)
            s_item.set_end_iteration_item(item)

    if len(start_iter_items) != 0:
        return False
    return True



"""
Created on Jul 25, 2021

@author: Benedikt Ursprung
"""
import pyqtgraph as pg
import h5py
from ScopeFoundry.data_browser import DataBrowserView

MEASURE_NAMES = ("auto_focus", "ranged_optimization")


class RangedOptimizationH5View(DataBrowserView):
    name = "ranged_optimization_h5"

    def is_file_supported(self, fname):
        # ADD other ranged optimization here
        for measure_name in MEASURE_NAMES:
            if measure_name + ".h5" in fname:
                self.measure_name = measure_name
                return True
        return False

    def setup(self):
        self.ui = plot_widget = pg.GraphicsLayoutWidget()

        self.axes = plot_widget.addPlot()
        self.axes.setLabel("left", "f(z)")
        self.axes.setLabel("bottom", "z")
        self.line_coarse = self.axes.plot(y=[0, 2, 1, 3, 2])
        self.line_fine = self.axes.plot(y=[0, 2, 1, 3, 2], pen="r")

        # indicator lines
        self.line_z_original = pg.InfiniteLine(
            movable=False,
            pen="b",
            label="original position: {value:0.6f}",
            labelOpts={
                "color": "b",
                "movable": True,
                "position": 0.15,
                "fill": (200, 200, 200, 200),
            },
        )
        self.line_z0_coarse = pg.InfiniteLine(
            movable=False,
            pen=(200, 200, 200),
            label="coarse optimized: {value:0.6f}",
            labelOpts={
                "color": (200, 200, 200),
                "movable": True,
                "position": 0.30,
                "fill": (200, 200, 200, 60),
            },
        )
        self.line_z0_fine = pg.InfiniteLine(
            movable=False,
            pen="r",
            label="fine optimized: {value:0.6f}",
            labelOpts={
                "color": "r",
                "movable": True,
                "position": 0.45,
                "fill": (200, 200, 200, 80),
            },
        )
        self.axes.addItem(self.line_z_original, ignoreBounds=True)
        self.axes.addItem(self.line_z0_coarse, ignoreBounds=True)
        self.axes.addItem(self.line_z0_fine, ignoreBounds=True)

    def on_change_data_filename(self, fname=None):
        with h5py.File(fname, "r") as file:
            group = file[f"measurement/{self.measure_name}"]

            self.axes.setTitle(self.measure_name)
            self.line_coarse.setData(group["z_coarse"][:], group["f_coarse"][:])
            self.line_z0_fine.setPos(group["z0_coarse"][()])
            self.line_z_original.setPos(group["z_original"][()])

            has_fine = bool(group["settings"].attrs["use_fine_optimization"])
            self.line_z0_fine.setVisible(has_fine)
            self.line_fine.setVisible(has_fine)
            if has_fine:
                self.line_fine.setData(group["z_fine"][:], group["f_fine"][:])
                self.line_z0_fine.setPos(group["z0_fine"][()])

            try:
                S = group["settings"].attrs
                self.axes.setLabel("left", S["f"])
                self.axes.setLabel("bottom", S["z_read"])
            except KeyError as e:
                print(e)



from typing_extensions import Type

factories = {}


from .base_item import BaseItem, VisitReturnType


def new_item(measure, item_type: str, **kwargs):
    return factories[item_type](measure, **kwargs)


def register_item(item: Type[BaseItem]):
    factories[item.item_type] = item


from .dir_operations import NewDir, NewDirEditorUI, SaveDirToParentEditorUI
from .exec_function import Function, ExecFunctionEditorUI
from .interrupt_if import InterruptIf, IterruptIfEditorUI
from .iterations import (
    EndIterationItem,
    InterationsEditorController,
    IterationsEditorUI,
    StartIterationItem,
    link_iteration_items,
)
from .pause import Pause, PauseEditorUI
from .read_from_hardware import ReadFromHardWare, ReadFromHardWareEditorUI
from .run_measurement import RunMeasurement, RunMeasurementEditorUI
from .timeout import Timeout, TimeoutEditorUI
from .update_settings import UpdateSetting, UpdateSettingEditorUI
from .wait_until import WaitUntil, WaitUntilEditorUI

register_item(InterruptIf)
register_item(NewDir)
register_item(Function)
register_item(StartIterationItem)
register_item(EndIterationItem)
register_item(Pause)
register_item(ReadFromHardWare)
register_item(RunMeasurement)
register_item(Timeout)
register_item(UpdateSetting)
register_item(WaitUntil)



from qtpy.QtWidgets import QComboBox, QSpinBox
from typing_extensions import TypedDict

from .base_item import BaseItem
from .editor_base_ui import EditorBaseUI
from .helper_func import new_q_completer

ITEM_TYPE = "measurement"


class RunMeasurementKwargs(TypedDict):
    measurement: str
    repetitions: int


class RunMeasurement(BaseItem):
    item_type = ITEM_TYPE

    def visit(self):
        m = self.app.measurements[self.kwargs["measurement"]]
        for i in range(self.kwargs["repetitions"]):
            try:
                self.measure.start_nested_measure_and_wait(m, nested_interrupt=False)
            except:
                print(f"{self.measure} delegated {m.name} failed")


class RunMeasurementEditorUI(EditorBaseUI):
    item_type = ITEM_TYPE
    description = "run a ScopeFoundry Measurement"

    def setup_ui(self):
        # # setting-update
        measure_layout = self.group_box.layout()
        measurements = self.measure.app.measurements.keys()
        self.measure_cb = QComboBox()
        self.measure_cb.setEditable(True)
        self.measure_cb.addItems(measurements)
        self.measure_cb.setCompleter(new_q_completer(measurements))
        measure_layout.addWidget(self.measure_cb)
        self.repetitions_sb = QSpinBox()
        self.repetitions_sb.setValue(1)
        self.repetitions_sb.setToolTip("number of repetitions")
        measure_layout.addWidget(self.repetitions_sb)

    def get_kwargs(self) -> RunMeasurementKwargs:
        k = self.measure_cb.currentText()
        reps = self.repetitions_sb.value()
        return {"measurement": k, "repetitions": reps}

    def set_kwargs(self, **kwargs):
        self.measure_cb.setCurrentText(kwargs["measurement"])
        self.repetitions_sb.setValue(kwargs["repetitions"])
        self.measure_cb.setFocus()



"""
Created on Sep 17, 2021

@author: Benedikt Ursprung
"""
from typing import Any, Dict, List, Union

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QListWidget

from .base_item import BaseItem


class ItemList:
    """implements the view, model and control of the list"""

    def __init__(self):
        self.view = QListWidget()
        self.view.setDefaultDropAction(Qt.MoveAction)
        self.view.setDragDropMode(QListWidget.DragDrop)

    def add(self, item: BaseItem, row: Union[int, None] = None):
        if row == None:
            row = self.get_current_row()
        self.view.insertItem(row + 1, item)
        self.view.setCurrentRow(row + 1)

    def remove(self, item: Union[BaseItem, None] = None):
        if item is not None:
            row = self.get_row(item)
        else:
            row = self.get_current_row()
        self.view.takeItem(row)
        del item

    def replace(self, new_item: BaseItem, old_item: Union[BaseItem, None] = None):
        if old_item is None:
            old_item = self.get_current_item()
        self.add(new_item, self.get_row(old_item))
        self.remove(old_item)

    def connect_item_double_clicked(self, fn):
        self.view.itemDoubleClicked.connect(fn)

    def get_view(self) -> QListWidget:
        return self.view

    def get_row(self, item: BaseItem) -> int:
        return self.view.row(item)

    def get_item(self, row: int) -> BaseItem:
        return self.view.item(row)  # type: ignore

    def get_current_row(self) -> int:
        return self.view.currentRow()

    def get_current_item(self) -> BaseItem:
        return self.view.currentItem()  # type: ignore

    def set_current_item(self, item: BaseItem):
        self.view.setCurrentItem(item)

    def clear(self):
        self.view.clear()

    def count_type(self, item_type="start-iteration"):
        counter = 0
        for i in range(self.view.count()):
            item = self.get_item(i)
            if item.item_type == item_type:
                counter += 1
        return counter

    def count(self) -> int:
        return self.view.count()

    def as_dicts(self) -> List[Dict[str, Any]]:
        l = []
        for i in range(self.view.count()):
            item = self.get_item(i)
            l.append({"type": item.item_type, **item.kwargs})
        return l



from qtpy.QtWidgets import QComboBox
from typing_extensions import TypedDict

from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem
from .helper_func import new_q_completer

ITEM_TYPE = "read_from_hardware"


class ReadFromHardWareKwargs(TypedDict):
    setting: str


class ReadFromHardWare(BaseItem):
    item_type = ITEM_TYPE

    def visit(self):
        self.app.get_lq(self.kwargs["setting"]).read_from_hardware()


class ReadFromHardWareEditorUI(EditorBaseUI):
    item_type = ITEM_TYPE
    description = "trigger read_from_hardware"

    def __init__(self, measure, paths) -> None:
        self.paths = paths
        super().__init__(measure)

    def setup_ui(self):
        self.setting_cb = QComboBox()
        self.setting_cb.setEditable(True)
        self.setting_cb.addItems(self.paths)
        self.setting_cb.setToolTip("setting to update")
        self.setting_cb.setCompleter(new_q_completer(self.paths))
        self.layout.addWidget(self.setting_cb)

    def get_kwargs(self) -> ReadFromHardWareKwargs:
        return {"setting": self.setting_cb.currentText()}

    def set_kwargs(self, **kwargs):
        self.setting_cb.setCurrentText(kwargs["setting"])
        self.setting_cb.setFocus()



"""
Created on Jan 23, 2023

@author: Benedikt Ursprung
"""
import time

import numpy
import pyqtgraph as pg
from qtpy import QtWidgets

from ScopeFoundry import Measurement, h5_io

INDEF = "indefinite"
TIMEOUT = "timeout"
ETOL = "error_tol"
ETOL_W_TIMEOUT = f"{ETOL}_with_{TIMEOUT}"


class PIDFeedbackControl(Measurement):
    name = "pid_feedback_control"

    def setup(self):
        s = self.settings
        s.New("Kp", float, initial=1)
        s.New("Ki", float, initial=0)
        s.New("Kd", float, initial=0)
        s.New("setpoint", float, initial=1.0, si=True)
        s.New(
            "sensor",
            str,
            initial="-",
            choices=("-",),
            description="choose a setting that supports read_from_hardware.",
        )
        s.New(
            "plant_input",
            str,
            initial="-",
            choices=("-",),
            description="choose a setting that has a hardware_set_func",
        )
        s.New(
            "timeout",
            float,
            vmin=0.0,
            initial=10.0,
            unit="s",
            description=f"used to define <b>terminate</b> condition <i>{TIMEOUT}</i> and <i>{ETOL_W_TIMEOUT}</i>, ignored otherwise",
        )
        s.New("error", float, initial=1_000, unit="%", ro=True)
        s.New(
            "error_tol",
            float,
            vmin=0.0,
            initial=1.0,
            unit="%",
            description=f"used to define <b>terminate</b> condition <i>{ETOL}</i> and <i>{ETOL_W_TIMEOUT}</i>, ignored otherwise",
        )
        s.New(
            "terminate",
            str,
            initial=ETOL_W_TIMEOUT,
            choices=(TIMEOUT, INDEF, ETOL_W_TIMEOUT, ETOL),
            description=f"condition for {self.name} to end.",
        )
        s.New("save_h5", bool, initial=False)
        self.data = mk_data_dict(600, 1.0)

    def setup_figure(self):
        s = self.settings

        ctr_layout = QtWidgets.QHBoxLayout()
        ctr_layout.addWidget(s.New_UI(("setpoint", "plant_input", "sensor")))
        ctr_layout.addWidget(s.New_UI(("Kp", "Ki", "Kd")))
        ctr_layout.addWidget(s.New_UI(("terminate", "timeout", "error_tol")))

        btn_layout = QtWidgets.QHBoxLayout()
        btn_layout.addWidget(s.get_lq("activation").new_pushButton())
        op_button = QtWidgets.QPushButton("unfilter options")
        op_button.clicked.connect(self.update_choices)
        btn_layout.addWidget(op_button)
        op_button = QtWidgets.QPushButton("filter options to available")
        op_button.setToolTip("only settings that are connected to hardware")
        op_button.clicked.connect(self.update_choices_filtered)
        btn_layout.addWidget(op_button)

        graph_widget = pg.GraphicsLayoutWidget(border=(0, 0, 0))
        axes = graph_widget.addPlot(title=self.name)
        t = self.data["lab_times"]
        self.sensor_values_plotline = axes.plot(t, self.data["sensor_values"], pen="w")
        self.setpoints_plotline = axes.plot(t, self.data["setpoints"], pen="y")

        self.ui = QtWidgets.QWidget()
        self.layout = QtWidgets.QVBoxLayout(self.ui)
        self.layout.addLayout(ctr_layout)
        self.layout.addLayout(btn_layout)
        self.layout.addWidget(graph_widget)

        self.update_choices()

    def run(self):
        from simple_pid.pid import PID  # pip install simple-pid (requires v>2)

        s = self.settings

        pid = PID(s["Kp"], s["Ki"], s["Kd"], setpoint=s["setpoint"])

        ii = 0
        wrap_around = 600
        self.data = mk_data_dict(wrap_around, s["setpoint"])
        t0 = time.time()
        lab_time = 0

        while not self.interrupt_measurement_called:
            pid.tunings = s["Kp"], s["Ki"], s["Kd"]
            pid.setpoint = s["setpoint"]
            sensor_value = self.app.get_lq(s["sensor"]).read_from_hardware()
            self.app.get_lq(s["plant_input"]).update_value(pid(sensor_value))
            s["error"] = 100.0 * (1 - sensor_value / s["setpoint"])
            lab_time = time.time() - t0

            self.data["setpoints"][ii] = s["setpoint"]
            self.data["sensor_values"][ii] = sensor_value
            self.data["lab_times"][ii] = lab_time

            if self.should_terminate(lab_time):
                break

            ii = (ii + 1) % wrap_around
            time.sleep(0.01)

        if s["save_h5"]:
            self.save_h5()

    def should_terminate(self, lab_time):
        s = self.settings
        if s["terminate"] == INDEF:
            return False
        if s["terminate"] == TIMEOUT:
            self.set_progress(100.0 * lab_time / s["timeout"])
            return lab_time >= s["timeout"]
        if s["terminate"] == ETOL:
            self.set_progress(100.0 * s["error_tol"] / abs(s["error"]))
            return abs(s["error"]) <= s["error_tol"]
        if s["terminate"] == ETOL_W_TIMEOUT:
            self.set_progress(100.0 * lab_time / s["timeout"])
            return lab_time >= s["timeout"] or abs(s["error"]) <= s["error_tol"]

    def update_display(self):
        t = self.data["lab_times"]
        self.sensor_values_plotline.setData(t, self.data["sensor_values"])
        self.setpoints_plotline.setData(t, self.data["setpoints"])

    def update_choices(self):
        s = self.settings
        s.get_lq("sensor").change_choice_list(self.app.get_setting_paths())
        s.get_lq("plant_input").change_choice_list(self.app.get_setting_paths())

    def update_choices_filtered(self):
        s = self.settings
        s.get_lq("sensor").change_choice_list(self.app.get_setting_paths(True, False))
        s.get_lq("plant_input").change_choice_list(self.app.get_setting_paths(False, True))

    def New_mini_UI(self):
        from qtpy.QtWidgets import QSizePolicy

        s = self.settings
        cb = s.get_lq("activation").new_default_widget()
        cb.setText(self.name)

        spw = s.get_lq("setpoint").new_default_widget()
        spw.setMaximumWidth(80)

        erw = s.get_lq("error").new_default_widget()
        erw.setMaximumWidth(80)

        show_ui_btn = QtWidgets.QPushButton("ui")
        show_ui_btn.setStyleSheet("QPushButton{font-weight: bold; color:blue};")
        show_ui_btn.setMaximumWidth(25)
        show_ui_btn.clicked.connect(self.show_ui)

        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(widget)
        layout.addWidget(cb)
        layout.addWidget(spw)
        layout.addWidget(erw)
        layout.addWidget(show_ui_btn)
        return widget

    def save_h5(self):
        h5_file = h5_io.h5_base_file(app=self.app, measurement=self)
        group = h5_io.h5_create_measurement_group(self, h5_file)
        for k, v in self.data.items():
            group[k] = v
        h5_file.close()


def mk_data_dict(N, setpoint):
    return {
        "setpoints": numpy.ones(N) * setpoint,
        "sensor_values": numpy.zeros(N),
        "lab_times": numpy.arange(N),
    }



'''
Created on Jul 23, 2014

Modified by Ed Barnard
UI enhancements by Ed Barnard, Alan Buckley
'''
from __future__ import print_function, division, absolute_import
from pathlib import Path

import sys
import time
import datetime
import numpy as np
import collections
from collections import OrderedDict
import logging
import inspect
from logging import Handler
import json

try:
    import configparser
except: # python 2
    import ConfigParser as configparser


from qtpy import QtCore, QtGui, QtWidgets
import pyqtgraph as pg
#import pyqtgraph.console

try:
    import IPython
    if IPython.version_info[0] < 4: #compatibility for IPython < 4.0 (pre Jupyter split)
        from IPython.qt.console.rich_ipython_widget import RichIPythonWidget as RichJupyterWidget
        from IPython.qt.inprocess import QtInProcessKernelManager
    else:
        from qtconsole.rich_jupyter_widget import RichJupyterWidget
        from qtconsole.inprocess import QtInProcessKernelManager
    CONSOLE_TYPE = 'qtconsole'
except Exception as err:
    logging.warning("ScopeFoundry unable to import iPython console, using pyqtgraph.console instead. Error: {}".format( err))
    import pyqtgraph.console
    CONSOLE_TYPE = 'pyqtgraph.console'
    
#import matplotlib
#matplotlib.rcParams['backend.qt4'] = 'PySide'
#from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
#from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT as NavigationToolbar2

#from matplotlib.figure import Figure

from .logged_quantity import LoggedQuantity, LQCollection
from .helper_funcs import confirm_on_close, ignore_on_close, load_qt_ui_file, \
    OrderedAttrDict, sibling_path, get_logger_from_class, str2bool
from . import h5_io, ini_io

#from equipment.image_display import ImageDisplay


import warnings
import traceback

# See https://riverbankcomputing.com/pipermail/pyqt/2016-March/037136.html
# makes sure that unhandled exceptions in slots don't crash the whole app with PyQt 5.5 and higher
# old version:
## sys.excepthook = traceback.print_exception
# new version to send to logger
def log_unhandled_exception(*exc_info):
    text = "".join(traceback.format_exception(*exc_info))
    logging.critical("Unhandled exception:" + text)
    #print("Unhandled exception:" + text)
sys.excepthook = log_unhandled_exception


# To fix a bug with jupyter qtconsole for python 3.8
# https://github.com/jupyter/notebook/issues/4613#issuecomment-548992047
import asyncio

if sys.platform == 'win32':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    
# Dark mode
try:
    import qdarktheme # pip install pyqtdarktheme
    darktheme_available = True
except Exception as err:
    darktheme_available = False
    print(f"pyqdarktheme unavailable: {err}")

class BaseApp(QtCore.QObject):
    
    def __init__(self, argv=[], dark_mode=False):
        QtCore.QObject.__init__(self)
        self.log = get_logger_from_class(self)
        
        path = Path(__file__)
        self.this_path = path.parent
        self.this_filename = path.name

        self.qtapp = QtWidgets.QApplication.instance()
        if not self.qtapp:
            self.qtapp = QtWidgets.QApplication(argv)
        
        if dark_mode and darktheme_available:
            qdarktheme.setup_theme()
        
        self.settings = LQCollection()
        
        # auto creation of console widget
        try:
            self.setup_console_widget()
        except Exception as err:
            print("failed to setup console widget " + str(err))
            self.console_widget = QtWidgets.QWidget()   

        
        # FIXME Breaks things for microscopes, but necessary for stand alone apps!
        #if hasattr(self, "setup"):
        #    self.setup()
        
        self.setup_logging()

        if not hasattr(self, 'name'):
            self.name = "ScopeFoundry"
        self.qtapp.setApplicationName(self.name)

        
    def exec_(self):
        return self.qtapp.exec_()
        
    def setup_console_widget(self, kernel=None):
        """
        Create and return console QWidget. If Jupyter / IPython is installed
        this widget will be a full-featured IPython console. If Jupyter is unavailable
        it will fallback to a pyqtgraph.console.ConsoleWidget.
        
        If the app is started in an Jupyter notebook, the console will be
        connected to the notebook's IPython kernel.
        
        the returned console_widget will also be accessible as self.console_widget
        
        In order to see the console widget, remember to insert it into an existing
        window or call self.console_widget.show() to create a new window      
        """
        if CONSOLE_TYPE == 'pyqtgraph.console':
            self.console_widget = pyqtgraph.console.ConsoleWidget(namespace={'app':self, 'pg':pg, 'np':np}, text="ScopeFoundry Console")
        elif CONSOLE_TYPE == 'qtconsole':
            
            if kernel == None:
                try: # try to find an existing kernel
                    #https://github.com/jupyter/notebook/blob/master/docs/source/examples/Notebook/Connecting%20with%20the%20Qt%20Console.ipynb
                    import ipykernel as kernel
                    conn_file = kernel.get_connection_file()
                    import qtconsole.qtconsoleapp
                    self.qtconsole_app = qtconsole.qtconsoleapp.JupyterQtConsoleApp()
                    self.console_widget = self.qtconsole_app.new_frontend_connection(conn_file)
                    self.console_widget.setWindowTitle("ScopeFoundry IPython Console")
                except: # make your own new in-process kernel
                    # https://github.com/ipython/ipython-in-depth/blob/master/examples/Embedding/inprocess_qtconsole.py
                    self.kernel_manager = QtInProcessKernelManager()
                    self.kernel_manager.start_kernel()
                    self.kernel = self.kernel_manager.kernel
                    self.kernel.shell.banner1 += """
                    ScopeFoundry Console
                    
                    Variables:
                     * np: numpy package
                     * app: the ScopeFoundry App object
                    """
                    self.kernel.gui = 'qt4'
                    self.kernel.shell.push({'np': np, 'app': self})
                    self.kernel_client = self.kernel_manager.client()
                    self.kernel_client.start_channels()
        
                    #self.console_widget = RichIPythonWidget()
                    self.console_widget = RichJupyterWidget()
                    self.console_widget.setWindowTitle("ScopeFoundry IPython Console")
                    self.console_widget.kernel_manager = self.kernel_manager
                    self.console_widget.kernel_client = self.kernel_client
            else:
                import qtconsole.qtconsoleapp
                self.qtconsole_app = qtconsole.qtconsoleapp.JupyterQtConsoleApp()
                self.console_widget = self.qtconsole_app.new_frontend_connection(kernel.get_connection_file())
                self.console_widget.setWindowTitle("ScopeFoundry IPython Console")
        else:
            raise ValueError("CONSOLE_TYPE undefined")
        
        return self.console_widget         

    def setup(self):
        pass


    def settings_save_ini(self, fname, save_ro=True):
        """"""
        config = configparser.ConfigParser()
        config.optionxform = str
        config.add_section('app')
        config.set('app', 'name', self.name)
        for lqname, lq in self.settings.as_dict().items():
            if not lq.ro or save_ro:
                config.set('app', lqname, lq.ini_string_value())
                
        with open(fname, 'w') as configfile:
            config.write(configfile)
        
        self.log.info("ini settings saved to {} {}".format( fname, config.optionxform))    

    def settings_load_ini(self, fname):
        self.log.info("ini settings loading from " + fname)
        

        config = configparser.ConfigParser()
        config.optionxform = str
        config.read(fname)

        if 'app' in config.sections():
            for lqname, new_val in config.items('app'):
                #print(lqname)
                lq = self.settings.as_dict().get(lqname)
                if lq:
                    if lq.dtype == bool:
                        new_val = str2bool(new_val)
                    lq.update_value(new_val)

    def settings_save_ini_ask(self, dir=None, save_ro=True):
        """Opens a Save dialogue asking the user to select a save destination and give the save file a filename. Saves settings to an .ini file."""
        # TODO add default directory, etc
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(self.ui, caption=u'Save Settings', dir=u"", filter=u"Settings (*.ini)")
        #print(repr(fname))
        if fname:
            self.settings_save_ini(fname, save_ro=save_ro)
        return fname

    def settings_load_ini_ask(self, dir=None):
        """Opens a Load dialogue asking the user which .ini file to load into our app settings. Loads settings from an .ini file."""
        # TODO add default directory, etc
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(None, "Settings (*.ini)")
        #print(repr(fname))
        if fname:
            self.settings_load_ini(fname)
        return fname  
    
    def setup_logging(self):
        
        logging.basicConfig(level=logging.WARN)#, filename='example.log', stream=sys.stdout)
        logging.getLogger('traitlets').setLevel(logging.WARN)
        logging.getLogger('ipykernel.inprocess').setLevel(logging.WARN)
        logging.getLogger('LoggedQuantity').setLevel(logging.WARN)
        logging.getLogger('PyQt5').setLevel(logging.WARN)
        logger = logging.getLogger('FoundryDataBrowser')
        
        self.logging_widget = QtWidgets.QWidget()
        self.logging_widget.setWindowTitle("Log")
        self.logging_widget.setLayout(QtWidgets.QVBoxLayout())
        self.logging_widget.search_lineEdit = QtWidgets.QLineEdit()
        self.logging_widget.log_textEdit = QtWidgets.QTextEdit("")
        
        self.logging_widget.layout().addWidget(self.logging_widget.search_lineEdit)
        self.logging_widget.layout().addWidget(self.logging_widget.log_textEdit)
        
        self.logging_widget.log_textEdit.document().setDefaultStyleSheet("body{font-family: Courier;}")
        
        self.logging_widget_handler = LoggingQTextEditHandler(
            self.logging_widget.log_textEdit, level=logging.DEBUG)
        logging.getLogger().addHandler(self.logging_widget_handler)
            
class LoggingQTextEditHandler(Handler, QtCore.QObject):
    
    new_log_signal = QtCore.Signal((str,))
    
    def __init__(self, textEdit, level=logging.NOTSET, buffer_len = 500):
        self.textEdit = textEdit
        self.buffer_len = buffer_len
        self.messages = []
        Handler.__init__(self, level=level)
        QtCore.QObject.__init__(self)
        self.new_log_signal.connect(self.on_new_log)

    def emit(self, record):
        log_entry = self.format(record)
        self.new_log_signal.emit(log_entry)
        
    def on_new_log(self, log_entry):
        #self.textEdit.moveCursor(QtGui.QTextCursor.End)
        #self.textEdit.insertHtml(log_entry)
        #self.textEdit.moveCursor(QtGui.QTextCursor.End)
        self.messages.append(log_entry)
        if len(self.messages) > self.buffer_len:
            self.messages = ["...<br>",] + self.messages[-self.buffer_len:]
        self.textEdit.setHtml("\n".join(self.messages))
        self.textEdit.moveCursor(QtGui.QTextCursor.End)
        
    level_styles = dict(
        CRITICAL="color: red;",
        ERROR="color: red;",
        WARNING='color: orange;',
        INFO='color: green;',
        DEBUG='color: green;',
        NOTSET='',
        )
    
    def format(self, record):
        #timestamp = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        timestamp = time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime(record.created))
        style = self.level_styles.get(record.levelname, "")        
        return """{} - <span style="{}">{}</span>: <i>{}</i> :{}<br>""".format(
            timestamp, style, record.levelname, record.name, record.msg)

class BaseMicroscopeApp(BaseApp):
    name = "ScopeFoundry"
    """The name of the microscope app, default is ScopeFoundry."""
    mdi = True
    """Multiple Document Interface flag. Tells the app whether to include an MDI widget in the app."""
    
    def __del__ ( self ): 
        self.ui = None

    def show(self):
        """Tells Qt to show the user interface"""
        #self.ui.exec_()
        self.ui.show()

    def __init__(self, argv=[], dark_mode=False):

        self._setting_paths = {}

        BaseApp.__init__(self, argv, dark_mode)
        
        log_path = Path.cwd() / 'log'
        if not log_path.is_dir():
            log_path.mkdir()
        log_fname = str(log_path / "{}_log_{:%y%m%d_%H%M%S}.txt".format(self.name, datetime.datetime.fromtimestamp(time.time())))
        self.log_file_handler = logging.FileHandler(log_fname)
        formatter = logging.Formatter('%(asctime)s|%(levelname)s|%(name)s|%(message)s', datefmt='%Y-%m-%dT%H:%M:%S')
        self.log_file_handler.setFormatter(formatter)

        logging.getLogger().addHandler(self.log_file_handler)
        
        initial_save_path = Path.cwd() / 'data'
        if not initial_save_path.is_dir():
            initial_save_path.mkdir()
        
        self.settings.New('save_dir', dtype='file', is_dir=True, initial=initial_save_path.as_posix())
        self.settings.New('sample', dtype=str, initial='')
        self.settings.New('data_fname_format', dtype=str,
                          initial='{timestamp:%y%m%d_%H%M%S}_{measurement.name}.{ext}')
                          # Potential new alternative default: '{unique_id_short}_{measurement.name}.{ext}'
        
        #self.settings.New('log_dir', dtype='file', is_dir=True, initial=initial_log_dir)
        
        if not hasattr(self, 'ui_filename'):
            if self.mdi:
                self.ui_filename = sibling_path(__file__,"base_microscope_app_mdi.ui")
            else:
                self.ui_filename = sibling_path(__file__,"base_microscope_app.ui")
        # Load Qt UI from .ui file
        self.ui = load_qt_ui_file(self.ui_filename)
        if self.mdi:
            self.ui.col_splitter.setStretchFactor(0,0)
            self.ui.col_splitter.setStretchFactor(1,1)
        
        self.hardware = OrderedAttrDict()
        self.measurements = OrderedAttrDict()

        self.quickbar = None
                   
        self.setup()
        
        self.setup_settings_paths()

        self.setup_default_ui()
        
        self.setup_ui()
      
        

    def setup_default_ui(self):
        self.ui.show()
        self.ui.activateWindow()
                
        """Loads various default features into the user interface upon app startup."""
        confirm_on_close(self.ui, title="Close %s?" % self.name, message="Do you wish to shut down %s?" % self.name, func_on_close=self.on_close)


        # Hardware and Measurement Settings Trees        
        self.ui.hardware_treeWidget.setColumnWidth(0,175)
        self.ui.measurements_treeWidget.setColumnWidth(0,175)

        self.ui.measurements_treeWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.ui.measurements_treeWidget.customContextMenuRequested.connect(self.on_measure_tree_context_menu)

        self.ui.hardware_treeWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.ui.hardware_treeWidget.customContextMenuRequested.connect(self.on_hardware_tree_context_menu)

        for name, hw in self.hardware.items():
            hw.add_widgets_to_tree(tree=self.ui.hardware_treeWidget)

        for name, measure in self.measurements.items():
            measure.add_widgets_to_tree(tree=self.ui.measurements_treeWidget)


        # Add log widget to mdiArea
        self.logging_subwin = self.add_mdi_subwin(self.logging_widget, "Log")
        self.console_subwin = self.add_mdi_subwin(self.console_widget, "Console")
        
        # Setup the Measurement UI's         
        for name, measure in self.measurements.items():
            self.log.info("setting up figures for {} measurement {}".format( name, measure.name) )            
            measure.setup_figure()
            if self.mdi and hasattr(measure, 'ui'):
                subwin = self.add_mdi_subwin(measure.ui, measure.name)
                measure.subwin = subwin
        
        if hasattr(self.ui, 'console_pushButton'):
            self.ui.console_pushButton.clicked.connect(self.console_widget.show)
            self.ui.console_pushButton.clicked.connect(self.console_widget.activateWindow)
                        
        if self.quickbar is None:
            # Collapse sidebar
            self.ui.quickaccess_scrollArea.setVisible(False)
        
        
        # Save Dir events
        self.ui.action_set_data_dir.triggered.connect(self.settings.save_dir.file_browser)
        self.settings.save_dir.connect_to_browse_widgets(self.ui.save_dir_lineEdit, self.ui.save_dir_browse_pushButton)
        
        # Sample meta data
        self.settings.sample.connect_bidir_to_widget(self.ui.sample_lineEdit)
        
        #settings button events
        if hasattr(self.ui, "settings_autosave_pushButton"):
            self.ui.settings_autosave_pushButton.clicked.connect(self.settings_auto_save_ini)
        if hasattr(self.ui, "settings_load_last_pushButton"):
            self.ui.settings_load_last_pushButton.clicked.connect(self.settings_load_last)
        if hasattr(self.ui, "settings_save_pushButton"):
            self.ui.settings_save_pushButton.clicked.connect(self.settings_save_dialog)
        if hasattr(self.ui, "settings_load_pushButton"):
            self.ui.settings_load_pushButton.clicked.connect(self.settings_load_dialog)
        
        #Menu bar entries:
        # TODO: connect self.ui.action_log_viewer to log viewer function
            # (Function has yet to be created)
        self.ui.action_load_ini.triggered.connect(self.settings_load_dialog)
        self.ui.action_auto_save_ini.triggered.connect(self.settings_auto_save_ini)
        self.ui.action_save_ini.triggered.connect(self.settings_save_dialog)
        self.ui.action_console.triggered.connect(self.console_widget.show)
        self.ui.action_console.triggered.connect(self.console_widget.activateWindow)
        self.ui.action_load_window_positions.triggered.connect(self.window_positions_load_dialog)
        self.ui.action_save_window_positions.triggered.connect(self.window_positions_save_dialog)
        
        #Refer to existing ui object:
        self.menubar = self.ui.menuWindow


        #Create new action group for switching between window and tab mode
        self.action_group = QtWidgets.QActionGroup(self)
        #Add actions to group:
        self.action_group.addAction(self.ui.window_action)
        self.action_group.addAction(self.ui.tab_action)
        
        self.ui.mdiArea.setTabsClosable(False)
        self.ui.mdiArea.setTabsMovable(True)
        
        self.ui.tab_action.triggered.connect(self.set_tab_mode)
        self.ui.window_action.triggered.connect(self.set_subwindow_mode)
        self.ui.cascade_action.triggered.connect(self.cascade_layout)
        self.ui.tile_action.triggered.connect(self.tile_layout)
        
        self.ui.setWindowTitle(self.name)

        # Set Icon
        logo_icon = QtGui.QIcon(sibling_path(__file__, "scopefoundry_logo2B_1024.png"))
        self.qtapp.setWindowIcon(logo_icon)
        self.ui.setWindowIcon(logo_icon)
        
        ### parameter tree
        ## disabled for now
        """
        import pyqtgraph.parametertree.parameterTypes as pTypes
        from pyqtgraph.parametertree import Parameter, ParameterTree, ParameterItem, registerParameterType
        
        self.ptree = ParameterTree()
        p = Parameter.create(name='Settings', type='group')
        
        app_params = Parameter.create(name='App', type='group')
        for lq_name, lq in self.settings.as_dict().items():
            print(lq_name, lq)
            lq_p = lq.new_pg_parameter()#Parameter.create(name=lq.name, type=lq.dtype)
            app_params.addChild(lq_p)
            
        p.addChild(app_params)
        
        

        hw_params = Parameter.create(name='Hardware', type='group')
        p.addChild(hw_params)
        
        for name, measure in self.hardware.items():
            hw_group = Parameter.create(name=name, type='group')
            hw_params.addChild(hw_group)
            for lq_name, lq in measure.settings.as_dict().items():
                print(lq_name, lq)
                lq_p = lq.new_pg_parameter()
                hw_group.addChild(lq_p)

        measure_params = Parameter.create(name='Measurements', type='group')
        p.addChild(measure_params)
        
        for name, measure in self.measurements.items():
            m_group = Parameter.create(name=name, type='group')
            measure_params.addChild(m_group)
            for lq_name, lq in measure.settings.as_dict().items():
                print(lq_name, lq)
                lq_p = lq.new_pg_parameter()
                m_group.addChild(lq_p)


        self.ptree.setParameters(p, showTop=True)
        #self.ptree.show()
        """
        
            
    def set_subwindow_mode(self):
        """Switches Multiple Document Interface to Subwindowed viewing mode."""
        self.ui.mdiArea.setViewMode(self.ui.mdiArea.SubWindowView)
    
    def set_tab_mode(self):
        """Switches Multiple Document Interface to Tabbed viewing mode."""
        self.ui.mdiArea.setViewMode(self.ui.mdiArea.TabbedView)
        
    def tile_layout(self):
        """Tiles subwindows in user interface. Specifically in the Multi Document Interface."""
        self.set_subwindow_mode()
        self.ui.mdiArea.tileSubWindows()
        
    def cascade_layout(self):
        """Cascades subwindows in user interface. Specifically in the Multi Document Interface."""
        self.set_subwindow_mode()
        self.ui.mdiArea.cascadeSubWindows()
        
    def bring_measure_ui_to_front(self, measure):
        self.bring_mdi_subwin_to_front(measure.subwin)
        
    def bring_mdi_subwin_to_front(self, subwin):
        viewMode = self.ui.mdiArea.viewMode()
        if viewMode == self.ui.mdiArea.SubWindowView:
            subwin.showNormal()
            subwin.raise_()
        elif viewMode == self.ui.mdiArea.TabbedView:
            subwin.showMaximized()
            subwin.raise_()
            
    def add_mdi_subwin(self, widget, name):
        subwin = self.ui.mdiArea.addSubWindow(widget, QtCore.Qt.CustomizeWindowHint | QtCore.Qt.WindowMinMaxButtonsHint)
        ignore_on_close(subwin)
        subwin.setWindowTitle(name)
        subwin.show()
        self.ui.menuWindow.addAction(name, lambda subwin=subwin: self.bring_mdi_subwin_to_front(subwin))
        return subwin
    
    def add_quickbar(self, widget):
        self.ui.quickaccess_scrollArea.setVisible(True)
        self.ui.quickaccess_scrollAreaWidgetContents.layout().addWidget(widget)
        self.quickbar = widget
        return self.quickbar
        
    def on_close(self):
        self.log.info("on_close")
        # disconnect all hardware objects
        for hw in self.hardware.values():
            self.log.info("disconnecting {}".format( hw.name))
            if hw.settings['connected']:
                try:
                    hw.disconnect()
                except Exception as err:
                    self.log.error("tried to disconnect {}: {}".format( hw.name, err) )

    def on_measure_tree_context_menu(self, position):
#         indexes =  self.ui.measurements_treeWidget.selectedIndexes()
#         if len(indexes) > 0:
#             level = 0
#             index = indexes[0]
#             while index.parent().isValid():
#                 index = index.parent()
#                 level += 1
#         if level == 0:
#             startAction = menu.addAction(self.tr("Start Measurement"))
#             interruptAction = menu.addAction(self.tr("Interrupt Measurement"))
        selected_items = self.ui.measurements_treeWidget.selectedItems()
        if len(selected_items) < 1:
            return
        selected_measurement_name = selected_items[0].text(0)
        if selected_measurement_name not in self.measurements:
            return
        M = self.measurements[selected_measurement_name]
        
        cmenu = QtWidgets.QMenu()        
        a = cmenu.addAction(selected_measurement_name)
        a.setEnabled(False)
        cmenu.addSeparator()
        cmenu.addAction("Start", M.start)
        cmenu.addAction("Interrupt", M.interrupt)
        cmenu.addSeparator()
        cmenu.addAction("Show", lambda M=M: self.bring_measure_ui_to_front(M))
        
        action = cmenu.exec_(QtGui.QCursor.pos())
    
    def on_hardware_tree_context_menu(self, position):
        selected_items = self.ui.hardware_treeWidget.selectedItems()
        if len(selected_items) < 1:
            return
        selected_hw_name = selected_items[0].text(0)
        if selected_hw_name not in self.hardware:
            return
        H = self.hardware[selected_hw_name]
        
        cmenu = QtWidgets.QMenu()        
        a = cmenu.addAction(selected_hw_name)
        a.setEnabled(False)
        connect_action = cmenu.addAction("Connect")
        disconnect_action = cmenu.addAction("Disconnect")
        
        action = cmenu.exec_(QtGui.QCursor.pos())
        if action == connect_action:
            H.settings['connected']=True
        elif action == disconnect_action:
            H.settings['connected']=False
        

    def setup(self):
        """ Override to add Hardware and Measurement Components"""
        #raise NotImplementedError()
        pass
    
        
    """def add_image_display(self,name,widget):
        print "---adding figure", name, widget
        if name in self.figs:
            return self.figs[name]
        else:
            disp=ImageDisplay(name,widget)
            self.figs[name]=disp
            return disp
    """
    
    def setup_ui(self):
        """ Override to set up ui elements after default ui is built"""
        pass
        
    def add_pg_graphics_layout(self, name, widget):
        self.log.info("---adding pg GraphicsLayout figure {} {}".format( name, widget))
        if name in self.figs:
            return self.figs[name]
        else:
            disp=pg.GraphicsLayoutWidget(border=(100,100,100))
            widget.layout().addWidget(disp)
            self.figs[name]=disp
            return disp
        
        # IDEA: write an abstract function to add pg.imageItem() for maps, 
        # which haddels, pixelscale, ROI ....
        # could also be implemented in the base_2d class? 
            
            
    
#     def add_figure_mpl(self,name, widget):
#         """creates a matplotlib figure attaches it to the qwidget specified
#         (widget needs to have a layout set (preferably verticalLayout) 
#         adds a figure to self.figs"""
#         print "---adding figure", name, widget
#         if name in self.figs:
#             return self.figs[name]
#         else:
#             fig = Figure()
#             fig.patch.set_facecolor('w')
#             canvas = FigureCanvas(fig)
#             nav    = NavigationToolbar2(canvas, self.ui)
#             widget.layout().addWidget(canvas)
#             widget.layout().addWidget(nav)
#             canvas.setFocusPolicy( QtCore.Qt.ClickFocus )
#             canvas.setFocus()
#             self.figs[name] = fig
#             return fig
    
    def add_figure(self,name,widget):
        # DEPRECATED
        return self.add_figure_mpl(name,widget)
    

    def add_hardware(self,hw):
        """Loads a HardwareComponent object into the app. 
        
        If *hw* is a class, rather an instance, create an instance 
        and add it to self.hardware
        """
        assert not hw.name in self.hardware.keys()

        #If *hw* is a class, rather an instance, create an instance 
        if inspect.isclass(hw):
            hw = hw(app=self)
        
        self.hardware.add(hw.name, hw)
                
        return hw
    
    
    def add_hardware_component(self,hw):
        # DEPRECATED use add_hardware()
        return self.add_hardware(hw)
    
    
    def add_measurement(self, measure):
        """Loads a Measurement object into the app.
        
        If *measure* is a class, rather an instance, create an instance 
        and add it to self.measurements

        """        
        #If *measure* is a class, rather an instance, create an instance 
        if inspect.isclass(measure):
            measure = measure(app=self)

        assert not measure.name in self.measurements.keys()
        
        self.measurements.add(measure.name, measure)

        return measure
    
    def add_measurement_component(self, measure):
        # DEPRECATED, use add_measurement()
        return self.add_measurement(measure)
    
    def settings_save_h5(self, fname):
        """
        Saves h5 file to a file.

        ==============  =========  =============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the h5 file.              
        ==============  =========  =============================================
        """
        with h5_io.h5_base_file(self, fname) as h5_file:
            for measurement in self.measurements.values():
                h5_io.h5_create_measurement_group(measurement, h5_file)
            self.log.info("settings saved to {}".format(h5_file.filename))
            
    def settings_save_ini(self, fname, save_ro=True, save_app=True, save_hardware=True, save_measurements=True):
        """
        ==============  =========  ==============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the ini file.              
        ==============  =========  ==============================================
        """
        exclude_patterns = []
        if not save_app:
            exclude_patterns.append("app")
        if not save_hardware:
            exclude_patterns.append("measurement")
        if not save_measurements:
            exclude_patterns.append("hardware")
        paths = self.get_setting_paths(exclude_patterns, exclude_ro = not save_ro)

        settings = self.read_settings(paths, ini_string_value=True)
        ini_io.save_settings(fname, settings)

        self.log.info(f"ini settings saved to {fname} str")

    def settings_load_ini(self, fname, ignore_hw_connect=False):
        """
        ==============  =========  ==============================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the ini file.              
        ==============  =========  ==============================================
        """
        settings = ini_io.load_settings(fname)
        if not ignore_hw_connect:
            self.write_settings_safe({k:v for k,v in settings.items() if k.endswith("connected")})
        self.write_settings_safe({k:v for k,v in settings.items() if not k.endswith("connected")})       

        
    def settings_load_h5(self, fname, ignore_hw_connect=False):
        """
        Loads h5 settings given a filename.

        ==============  =========  ====================================================================================
        **Arguments:**  **Type:**  **Description:**
        fname           str        relative path to the filename of the h5 file.              
        ==============  =========  ====================================================================================
        """
        settings = h5_io.load_settings(fname)
        if not ignore_hw_connect:
            self.write_settings_safe({k:v for k,v in settings.items() if k.endswith("connected")})
        self.write_settings_safe({k:v for k,v in settings.items() if not k.endswith("connected")})
    
    def settings_auto_save_ini(self):
        """
        Saves the ini file to app/save_dir directory with a time stamp in the filename.
        """
        fname = Path(self.settings["save_dir"]) / f"{datetime.datetime.now():%y%m%d_%H%M%S}_settings.ini"
        self.settings_save_ini(fname)

    def settings_load_last(self):
        """
        Loads last saved ini file.
        """
        fnames = Path.cwd().glob("*_settings.ini")
        fnames.extend( Path(self.settings["save_dir"]).glob("*_settings.ini"))
        self.settings_load_ini(sorted(fnames)[-1])
    
    
    def settings_save_dialog(self):
        """Opens a save as ini dialogue in the app user interface."""
        fname, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(self.ui, "Save Settings file", "", "Settings File (*.ini)")
        if fname:
            self.settings_save_ini(fname)
    
    def settings_load_dialog(self):
        """Opens a load ini dialogue in the app user interface"""
        fname, selectedFilter = QtWidgets.QFileDialog.getOpenFileName(self.ui,"Open Settings file", "", "Settings File (*.ini *.h5)")
        if fname.endswith(".ini"):
            self.settings_load_ini(fname)
        elif fname.endswith(".h5"):
            self.settings_load_h5(fname)
        
    def window_positions_load_dialog(self):
        fname, selectedFilter = QtWidgets.QFileDialog.getOpenFileName(self.ui,"Open Window Position file", "", "position File (*.json)")
        self.load_window_positions_json(fname)
        
    def window_positions_save_dialog(self):
        """Opens a save as ini dialogue in the app user interface."""
        fname, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(self.ui, "Save Window Position file", "", "position File (*.json)")
        if fname:
            self.save_window_positions_json(fname)
        
    def get_lq(self, path:str) -> LoggedQuantity:
        """
        returns the LoggedQuantity defined by a path string of the form 'section/[component/]setting'
        where section are "mm", "hw" or "app"
        """
        parts = path.split("/")
        section = parts[0]
        if section in ("HW", "hardware"):
            path = f"hw/{parts[1]}/{parts[2]}"
        elif section in ("measurement", "measure", "measurements"):
            path = f"mm/{parts[1]}/{parts[2]}"
        if not path in self._setting_paths:
            print(f"WARNING: {'/'.join(parts)} does not exist")
        else:
            return self._setting_paths[path]

    def write_setting(self, path:str, value):
        self.get_lq(path).update_value(value)

    def write_setting_safe(self, path:str, value):
        lq = self.get_lq(path)
        if lq is None or lq.protected:
            return
        lq.update_value(value)      

    def write_settings_safe(self, settings):
        """
        updates settings based on a dictionary, silently ignores protected logged quantities and non-existing.  

        ==============  =========  ====================================================================================
        **Arguments:**  **Type:**  **Description:**
        settings        dict       (path, value) map
        ==============  =========  ====================================================================================
        """
        for path, value in settings.items():
            self.write_setting_safe(path, value)

    def read_setting(self, path:str, read_from_hardware=True, ini_string_value=False):
        lq = self.get_lq(path)
        if read_from_hardware and lq.has_hardware_read:
            lq.read_from_hardware()
        if ini_string_value:
            return lq.ini_string_value()
        return lq.val

    def setup_settings_paths(self):
        for hw_name, hw in self.hardware.items():
            for name, lq in hw.settings.as_dict().items():
                self.add_setting_path(f"hw/{hw_name}/{name}", lq)
        for mm_name, mm in self.measurements.items():
            for name, lq in mm.settings.as_dict().items():
                self.add_setting_path(f"mm/{mm_name}/{name}", lq)
        for name, lq in self.settings.as_dict().items():
            self.add_setting_path(f"app/{name}", lq)

    def add_setting_path(self, path:str, lq: LoggedQuantity):
        lq.set_path(path)
        self._setting_paths[path] = lq

    def get_setting_paths(self, filter_has_hardware_read=False, filter_has_hardware_write=False, exclude_patterns=None, exclude_ro=False):        
        if filter_has_hardware_read and filter_has_hardware_write:
            paths = (path for path, lq in self._setting_paths.items() if lq.has_hardware_read() or lq.has_hardware_write())
        elif filter_has_hardware_read:
            paths = (path for path, lq in self._setting_paths.items() if lq.has_hardware_read())
        elif filter_has_hardware_write:
            paths = (path for path, lq in self._setting_paths.items() if lq.has_hardware_write())
        else:
            paths = self._setting_paths.keys()
        if exclude_ro:
            ro_paths = [path for path in paths if self.get_lq(path).ro]
            exclude_patterns = ro_paths if not exclude_patterns else list(exclude_patterns) + ro_paths
        if exclude_patterns:
            paths = (path for path in paths if not any(pattern in path for pattern in exclude_patterns))        
        return list(paths)
    
    def read_settings(self, paths=None, read_from_hardware=False, ini_string_value=False):
        """returns a dictionary (path, value):
        ================== =========  =============================================================================
        **Arguments:**     **Type:**  **Description:**
        paths              list[str]  paths to setting, if None(default) all paths are used
        read_from_hardware bool       if True, values are read from hardware, else the current value is used
        ================== =========  =============================================================================
        """
        paths = self.get_setting_paths() if paths is None else paths
        return {p:self.read_setting(p, read_from_hardware, ini_string_value) for p in paths}

    def lq_path(self, path):
        warnings.warn("App.lq_path deprecated, use App.get_lq instead", DeprecationWarning)
        return self.get_lq(path)

    def lq_paths_list(self):
        warnings.warn("App.lq_paths_list deprecated, use App.get_setting_paths instead", DeprecationWarning)
        return self.get_setting_paths()
        
    @property
    def hardware_components(self):
        warnings.warn("App.hardware_components deprecated, used App.hardware", DeprecationWarning)
        return self.hardware

    @property
    def measurement_components(self):
        warnings.warn("App.measurement_components deprecated, used App.measurements", DeprecationWarning)
        return self.measurements
    
    @property
    def logged_quantities(self):
        warnings.warn('app.logged_quantities deprecated use app.settings', DeprecationWarning)
        return self.settings.as_dict()
    
    def set_window_positions(self, positions):
        def restore_win_state(subwin, win_state):
            subwin.showNormal()
            if win_state['maximized']:
                subwin.showMaximized()
            elif win_state['minimized']:
                subwin.showMinimized()
            else:
                subwin.setGeometry(*win_state['geometry'])
            
        self.set_subwindow_mode()
        for name, win_state in positions.items():
            if name == 'log':
                restore_win_state(self.logging_subwin, win_state)
            elif name == 'console':
                restore_win_state(self.console_subwin, win_state)
            elif name == 'main':
                restore_win_state(self.ui, win_state)
                self.ui.col_splitter.setSizes(win_state['col_splitter_sizes'])
            elif name.startswith('measurement/'):
                M = self.measurements[name.split('/')[-1]]
                restore_win_state(M.subwin, win_state)
            
        
        
    def get_window_positions(self):
        positions = OrderedDict()
        
        def qrect_to_tuple(qr):
            return  (qr.x(), qr.y(), qr.width(), qr.height())
        
        def win_state_from_subwin(subwin):
            window_state = dict(
                    geometry  = qrect_to_tuple(subwin.geometry()),
                    maximized = subwin.isMaximized(),
                    minimized = subwin.isMinimized(),
                    fullscreen = subwin.isFullScreen()
                    )
            return window_state
        
        positions['main'] = win_state_from_subwin(self.ui)
        positions['main']['col_splitter_sizes'] = self.ui.col_splitter.sizes()
            
        positions['log'] = win_state_from_subwin(self.logging_subwin)
        positions['console'] = win_state_from_subwin(self.console_subwin)

        for name, M in self.measurements.items():
            if hasattr(M, 'ui'):
                positions['measurement/'+name] = win_state_from_subwin(M.subwin)
       
        return positions
    
    def save_window_positions_json(self, fname):
        positions = self.get_window_positions()
        with open(fname, 'w') as outfile:    
            json.dump(positions, outfile, indent=4)
            
    def load_window_positions_json(self, fname):
        with open(fname, 'r') as infile:
            positions = json.load(infile)
        self.set_window_positions(positions)
        
#     def save_window_positions_ini(self, fname):
#         """
#         ==============  =========  ==============================================
#         **Arguments:**  **Type:**  **Description:**
#         fname           str        relative path to the filename of the ini file.              
#         ==============  =========  ==============================================
#         """
#         positions = self.get_window_positions()
# 
#         config = configparser.ConfigParser(interpolation=None)
#         config.optionxform = str
#         
#         for name, win_state in positions.items():
#             config.add_section(name)
#             for k, v in win_state.items():
#                 config.set(name, k, v)
#         with open(fname, 'w') as configfile:
#             config.write(configfile)
#         
#         self.log.info("ini windown settings saved to {} {}".format( fname, config.optionxform))

    def generate_data_path(self, measurement, ext,t=None):
        if t is None:
            t = time.time()
        f = self.settings['data_fname_format'].format(
            app=self,
            measurement=measurement,
            timestamp=datetime.datetime.fromtimestamp(t),
            ext=ext)
        return Path(self.settings['save_dir']) / f



if __name__ == '__main__':
    
    app = BaseMicroscopeApp(sys.argv)
    
    sys.exit(app.exec_())


import time
from datetime import datetime
from pathlib import Path

from qtpy.QtWidgets import QLabel, QLineEdit
from typing_extensions import TypedDict

from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem

ITEMTYPE = "new_dir"
SAVEDIRITEMTYPE = "save_dir_to_parent"


class NewDirKwargs(TypedDict):
    new_dir_name: str


class NewDir(BaseItem):
    item_type = ITEMTYPE

    def visit(self):
        t0 = time.time()
        name = self.kwargs["new_dir_name"]
        sub_dir = f"{datetime.fromtimestamp(t0):%y%m%d_%H%M%S}_{name}"
        new_dir = Path(self.app.settings["save_dir"]) / sub_dir
        new_dir.mkdir()
        self.app.settings["save_dir"] = new_dir.as_posix()


class NewDirEditorUI(EditorBaseUI):
    item_type = ITEMTYPE
    description = f"creates sub folder and set as save_dir"

    def setup_ui(self):
        self.new_dir_name_lineEdit = QLineEdit()
        self.group_box.layout().addWidget(self.new_dir_name_lineEdit)

    def get_kwargs(self):
        val = self.new_dir_name_lineEdit.text()
        return {"new_dir_name": val}

    def set_kwargs(self, **kwargs):
        self.new_dir_name_lineEdit.setText(kwargs["new_dir_name"])


class SaveDirToParent(BaseItem):
    item_type = SAVEDIRITEMTYPE

    def visit(self):
        cur = Path(self.app.settings["save_dir"])
        self.app.settings["save_dir"] = cur.parent.as_posix()


class SaveDirToParentKwargs(TypedDict):
    info: str


class SaveDirToParentEditorUI(EditorBaseUI):
    item_type = SAVEDIRITEMTYPE
    description = "save_dir to parent, designed use in conjunction with new_dir"

    def setup_ui(self):
        self.spacer = QLabel()
        self.layout.addWidget(self.spacer)

    def get_kwargs(self) -> SaveDirToParentKwargs:
        return {"info": "save_dir jumps to parent"}

    def set_kwargs(self, **kwargs):
        self.spacer.setFocus()



from typing import List
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QCompleter


def new_q_completer(l: List[str]) -> QCompleter:
    completer = QCompleter(l)
    completer.setCompletionMode(QCompleter.PopupCompletion)
    completer.setModelSorting(QCompleter.UnsortedModel)
    completer.setFilterMode(Qt.MatchContains)
    completer.setCaseSensitivity(Qt.CaseInsensitive)
    return completer



from __future__ import division, print_function, absolute_import
from ScopeFoundry import HardwareComponent
import random

class DummmyXYStageDevice(object):
    
    def __init__(self, debug=False):
        self.x = 0
        self.y = 0
        self.debug=debug
        # communicate with hardware here
    
    def read_x(self):
        self.x = self.x + self.noise()
        #if self.debug: print "read_x", self.x
        #print("read_x", self.x)
        return self.x

    def read_y(self):
        self.y = self.y + self.noise()
        #if self.debug: print "read_y", self.y
        return self.y
    
    def write_x(self, x):
        self.x = x
        #print("write_x", self.x, x)
        #if self.debug: print "write_x", self.x
        

    def write_y(self, y):
        self.y = y
        #if self.debug: print "write_y", self.y
    
    def close(self):
        #print "dummy_xy_stage_equipment close"
        pass
    
    def noise(self):
        return (random.random()-0.5)*10e-3

class DummyXYStageHW(HardwareComponent):
    
    name = "dummy_xy_stage"
    
    def setup(self):
        lq_params = dict(  dtype=float, ro=False,
                           initial = -1,
                           vmin=-1,
                           vmax=100,
                           si = False,
                           unit='um')
        self.x_position = self.add_logged_quantity("x_position", **lq_params)
        self.y_position = self.add_logged_quantity("y_position", **lq_params)       

        self.x_position.reread_from_hardware_after_write = True
        self.x_position.spinbox_decimals = 3
        
        self.y_position.reread_from_hardware_after_write = True
        self.y_position.spinbox_decimals = 3

    def connect(self):
        #if self.debug_mode.val: print "connecting to dummy_xy_stage"

        # Open connection to hardware
        self.stage_device = DummmyXYStageDevice(debug=self.debug_mode.val)

        # connect logged quantities
        self.x_position.hardware_read_func = self.stage_device.read_x
        self.y_position.hardware_read_func = self.stage_device.read_y
        
        self.x_position.hardware_set_func  = self.stage_device.write_x
        self.y_position.hardware_set_func  = self.stage_device.write_y

    def disconnect(self):
        #if self.debug_mode.val: print "disconnecting to dummy_xy_stage"
        
        #disconnect logged quantities from hardware
        for lq in self.settings.as_list():
            lq.hardware_read_func = None
            lq.hardware_set_func = None
        
        if hasattr(self, 'stage_device'):
            #disconnect hardware
            self.stage_device.close()
            
            # clean up hardware object
            del self.stage_device






from qtpy.QtWidgets import QGroupBox, QHBoxLayout, QPushButton


class EditorBaseUI:
    item_type = ""
    description = ""

    def __init__(self, measure) -> None:
        self.measure = measure

        self.layout = layout = QHBoxLayout()
        self.group_box = gb = QGroupBox(self.item_type.replace("_", "-"))
        gb.setToolTip(self.description)
        gb.setLayout(layout)

        self.add_btn = add_btn = QPushButton("New")
        add_btn.setToolTip("CTRL+N")
        add_btn.setFixedWidth(30)
        layout.addWidget(add_btn)

        self.replace_btn = replace_btn = QPushButton("Replace")
        replace_btn.setFixedWidth(50)
        replace_btn.setToolTip("CTRL+R")
        layout.addWidget(replace_btn)

        self.setup_ui()

    def set_on_new_func(self, fn):
        self.add_btn.clicked.connect(fn)

    def set_on_replace_func(self, fn):
        self.replace_btn.clicked.connect(fn)

    def setup_ui(self):
        ...

    def get_kwargs(self):
        ...

    def set_kwargs(self, **kwargs):
        ...



from qtpy.QtWidgets import (QCheckBox, QComboBox, QGridLayout, QGroupBox,
                            QHBoxLayout, QLabel, QLineEdit, QPushButton,
                            QTextEdit, QVBoxLayout)


class PublishOnGhView(QGroupBox):

    def __init__(self) -> None:
        super().__init__()
        self.setup()

    def setup(self) -> None:
        layout = QVBoxLayout()
        self.setLayout(layout)

        self.my_gh_acc_label = QLabel('my GitHub')
        self.my_gh_acc_le = QLineEdit('UBene')
        self.my_gh_acc_validate_btn = QPushButton('validate')
        self.my_gh_acc_validate_btn.setMaximumWidth(55)
        self.my_gh_acc_is_valid_cb = QCheckBox('')
        self.my_gh_acc_is_valid_cb.setEnabled(False)
        self.my_gh_acc_is_valid_cb.setMaximumWidth(15)
        my_gh_layout = QHBoxLayout()
        my_gh_layout.addWidget(self.my_gh_acc_label)
        my_gh_layout.addWidget(self.my_gh_acc_le)
        layout.addLayout(my_gh_layout)

        self.hw_label = QLabel('Hardware')
        self.hw_cb = QComboBox()
        self.update_hws = QPushButton('update')
        self.update_hws.setMaximumWidth(55)
        hw_layout = QHBoxLayout()
        hw_layout.addWidget(self.hw_label)
        hw_layout.addWidget(self.hw_cb)
        hw_layout.addWidget(self.update_hws)
        layout.addLayout(hw_layout)

        publish_layout = QHBoxLayout()
        self.subtree_btn = QPushButton('1. subtree push')
        publish_layout.addWidget(self.subtree_btn)
        self.publish_btn = QPushButton('2. publish')
        publish_layout.addWidget(self.publish_btn)

        layout.addLayout(publish_layout)

        self.info_box = QTextEdit('')
        layout.addWidget(self.info_box)

    def update_hw_options_clicked(self, callback):
        self.update_hws.clicked.connect(callback)

    def validate_my_gh_acc_clicked(self, callback):
        self.my_gh_acc_validate_btn.clicked.connect(callback)

    def publish_clicked(self, callback):
        self.publish_btn.clicked.connect(callback)

    def subtree_clicked(self, callback):
        self.subtree_btn.clicked.connect(callback)

    def set_my_gh_acc_valid(self, checked):
        self.my_gh_acc_is_valid_cb.setChecked(checked)

    def get_hw(self) -> str:
        return self.hw_cb.currentText()

    def set_hw_options(self, options):
        self.hw_cb.clear()
        self.hw_cb.addItems(options)

    def get_my_gh_acc(self) -> str:
        return self.my_gh_acc_le.text()

    def set_my_gh_acc(self, text):
        self.my_gh_acc_le.setText(text)

    def set_info_text(self, text):
        self.info_box.setHtml(text)

    def append_info(self, text):
        self.info_box.setHtml(self.info_box.toHtml() + text)



'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''

# INFO: GOAL OF THIS FILE IS TO HANDLE LOW LEVEL COMMUNICATION THROUGH A DLL
# SHOULD BE INDEPENDENT OF ANY SCOPEFOUNDRY FUNCTIONALITY

import ctypes
from threading import Lock


class $DEV_CLASS_NAME:

    def __init__(self, dll_path=r"path_to_dll", debug=False):
        """
        serial_num if defined should be a string or integer
        """
        self.debug = debug
        self.dll = ctypes.windll.LoadLibrary(dll_path + r"/XX.dll")
        self.lock = Lock()
        self._id = self.open()

    def open(self):
        raise NotImplementedError
        # assuming dll .h file has init()
        # self.dll.init()
        #
        # Assuming dll .h file has
        #   void OpenFirst(int *id)
        # _id = ctypes.c_int()
        # with self.lock:
        #     self.dll.OpenFirst(ctypes.byref(_id))
        # return _id  #typically can be left as a ctype

    def close(self):
        raise NotImplementedError
        # good pratice to implement a close function
        # find CLOSE function in dll header file
        # with self.lock:
        #     self.dll.CLOSE(self._id)

    def write_property_x(self, value):
        raise NotImplementedError

        # assuming header file says
        #   void SET_PROPERTY_X(_id, float value)
        # with self.lock:
        #     self.dll.SET_PROPERTY_X(ctypes.c_float(value))
        # if self.debug:
        #     print(write_property_x to value)
        # find CLOSE function in dll header file

    def read_property_x(self):
        raise NotImplementedError

        # assuming header file says
        #   void GET_PROPERTY_X(int id, float *value)
        # value = ctypes.c_float()
        # with self.lock:
        #     self.dll.SET_PROPERTY_X(self._id, ctypes.byref(cvalue))
        # if self.debug:
        #     print(cvalue.value)
        # return cvalue.value

    def read_data(self):
        raise NotImplementedError
        # function name to generic, be more specific
        #
        # return a list of data points


if __name__ == '__main__':
    dev = $DEV_CLASS_NAME('path_to_dll', debug=True)
    print(dev.read_property_x())
    dev.write_property_x(999.1)
    print(dev.read_property_x())
    dev.close()



from .item_list import ItemList
from .editor_base_ui import EditorBaseUI
from . import new_item


class EditorBaseController:
    def __init__(self, editor_ui: EditorBaseUI, measure) -> None:
        self.ui = editor_ui
        self.item_type = editor_ui.item_type
        self.description = editor_ui.description
        self.ui.set_on_new_func(self.on_new_func)
        self.ui.set_on_replace_func(self.on_replace_func)
        self.measure = measure
        self.item_list: ItemList = measure.item_list

    def _new_item(self):
        return new_item(self.measure, self.item_type, **self.ui.get_kwargs())

    def on_new_func(self):
        self.item_list.add(self._new_item())

    def on_replace_func(self):
        self.item_list.replace(self._new_item())



from ScopeFoundry import Measurement
from ScopeFoundry.helper_funcs import sibling_path, load_qt_ui_file
import configparser
import time
from ScopeFoundry.logged_quantity import LQCollection, LoggedQuantity
from collections import OrderedDict

class BaseRecipeControl(Measurement):
    """
    Load and save Recipes / settings
    """
    
    name = 'recipe_control'
    
    def __init__(self, app, name=None, settings_dict=dict(sample='app/sample'), ):
        
        self.settings_dict = settings_dict

        Measurement.__init__(self, app, name=name)


    def setup(self):
        
        self.settings.New('recipes_filename', dtype='file', initial='')
        self.settings.recipes_filename.add_listener(self.load_recipes_file)

        self.settings.New('recipe_name', dtype=str, initial='recipe1', choices=('recipe1', ))
        self.settings.New('recipe_date_modified', dtype=str, ro=True)

        #self.recipe_settings.Add(self.settings.recipe_name)
        #self.recipe_settings.Add(self.settings.recipe_date_modified)


        for name, lq_path in self.settings_dict.items():
            if isinstance(lq_path, LoggedQuantity):
                lq = lq_path
            else:
                lq = self.app.lq_path(lq_path)
            recipe_lq = self.settings.New('recipe_' + name,
                              dtype=lq.dtype,
                              unit=lq.unit,
                              choices=lq.choices,
                              spinbox_decimals=lq.spinbox_decimals,
                              ro=True)
            
            self.settings_dict[name] = (recipe_lq, lq)            
            
        # list of recipe dicts
        self.recipes = []        
    
    def setup_figure(self):
        self.load_default_recipe_ui()

    def load_default_recipe_ui(self):
        self.ui = load_qt_ui_file(sibling_path(__file__, 'stored_recipe_control.ui'))

        for name, (recipe_lq, lq) in self.settings_dict.items():
            self.ui.recipe_settings_groupBox.layout().addRow(name, recipe_lq.new_default_widget())
            self.ui.current_settings_groupBox.layout().addRow(name, lq.new_default_widget())

        self.settings.recipe_date_modified.connect_to_widget(self.ui.recipe_date_modified_label)

        self.ui.save_recipe_pushButton.clicked.connect(self.on_save_recipe)
        self.ui.execute_pushButton.clicked.connect(self.execute_current_recipe)
        self.ui.delete_recipe_pushButton.clicked.connect(self.delete_current_recipe)
        
        self.settings.recipe_name.connect_to_widget(self.ui.recipe_name_comboBox)
        self.settings.recipe_name.add_listener(self.select_current_recipe)
        

        
        self.settings.recipes_filename.connect_to_browse_widgets(
            self.ui.recipes_filename_lineEdit,
            self.ui.recipes_filename_browse_pushButton)
        



    def get_recipe_by_name(self, name):
        for recipe in self.recipes:
            if recipe['name'] == name:
                return recipe
        raise ValueError("recipe not found {}".format(name))


    def load_recipes_file(self):
        
        config = configparser.ConfigParser()
        config.optionxform = str
        config.read(self.settings['recipes_filename'])

        self.recipes.clear()
        for name in config.sections():
            recipe_dict = OrderedDict()
            recipe_dict['name'] = name
            for key, val in config.items(name):
                recipe_dict[key] = val
            self.recipes.append(recipe_dict)
    
        # update recipe choices
        self.settings.recipe_name.change_choice_list(tuple([r['name'] for r in self.recipes]))
        
        # set recipe_name to first record if current recipe_name is not in file
        current_recipe_name = self.settings['recipe_name']
        if current_recipe_name in self.recipes:
            self.select_current_recipe(current_recipe_name)
        else:
            self.settings['recipe_name'] = self.recipes[0]['name']

    def save_recipes_file(self):
        config = configparser.ConfigParser()
        config.optionxform = str
        
        for recipe in self.recipes:
            config.add_section(recipe['name'])
            for setting_name in (list(self.settings_dict.keys()) + ['date_modified',]):
                if setting_name in recipe:
                    print(recipe['name'], setting_name, recipe[setting_name])
                    config.set(recipe['name'], setting_name, str(recipe[setting_name]))
    
        with open(self.settings['recipes_filename'], 'w') as configfile:
            config.write(configfile)



    def select_current_recipe(self, name=None):
        print("select_current_recipe", name)
        if name is None:
            name = self.settings['recipe_name']
        print("select_current_recipe", name)
        recipe = self.get_recipe_by_name(name)
        
        self.ui.new_recipe_name_lineEdit.setText(name)       
        # update recipe logged quantities
        for setting_name in self.settings_dict.keys():
            self.settings['recipe_' + setting_name] = recipe[setting_name]
            
        self.settings['recipe_date_modified'] = recipe['date_modified']
    
    
    def delete_current_recipe(self):
        recipe = self.get_recipe_by_name(self.settings['recipe_name'])        
        self.recipes.remove(recipe) 
        self.save_recipes_file()
        self.load_recipes_file()
    
    
    def save_current_settings_as_recipe(self, name):
        new_recipe = OrderedDict()
        new_recipe['name'] = name
        for setting_name, (recipe_lq, current_lq) in self.settings_dict.items():
            new_recipe[setting_name] = current_lq.value
        new_recipe['date_modified'] = time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime(time.time()))
        
        try:
            old_recipe = self.get_recipe_by_name(name)
            if old_recipe['date_modified'] == 'SYSTEM':
                self.load_recipes_file()
                self.settings['recipe_name'] = name

                return
            old_recipe.update(new_recipe)
        except ValueError:
            self.recipes.append(new_recipe)

        self.save_recipes_file()
        self.load_recipes_file()
        self.settings['recipe_name'] = name
        
    
    def execute_current_recipe(self):
        # save first?
        # ask first?
        for setting_name, (recipe_lq, current_lq) in self.settings_dict.items():
            current_lq.update_value(recipe_lq.value)

    
    
    def on_save_recipe(self):
        new_name = self.ui.new_recipe_name_lineEdit.text()
        self.save_current_settings_as_recipe(new_name)



if __name__ == '__main__':
    from ScopeFoundry.base_app import BaseMicroscopeApp
    
    class TestApp(BaseMicroscopeApp):
        name = 'test_app'
        
        def setup(self):
            self.settings.New('asdf', dtype=float, initial=4)
            settings_dict = OrderedDict()
            settings_dict['asdf'] = 'app/asdf'
            settings_dict['sample'] = self.settings.sample
            M = self.add_measurement(BaseRecipeControl(self, settings_dict=settings_dict))
            
            M.settings['recipes_filename'] = 'test_stored_recipe_control.ini'

    
    app = TestApp([])
    app.exec_()


"""helpers for handling .ini files"""
import configparser

TRANSLATIONS = {"True": True, "False": False}
REV_TRANSLATIONS = {v: k for k, v in TRANSLATIONS.items()}


def to_path_settings(config):
    settings = {}
    for section, values in config.items():
        for name, val in values.items():
            settings[f"{section}/{name}"] = TRANSLATIONS.get(val, val)
    return settings


def from_path_settings(settings):
    config = configparser.ConfigParser(interpolation=None)
    config.optionxform = str

    for path, value in settings.items():
        parts = path.split("/")
        section = "/".join(parts[:-1])
        name = parts[-1]
        if not section in config.sections():
            config.add_section(section)
        config.set(section, name, REV_TRANSLATIONS.get(value, value))

    return config


def laod_ini(fname):
    config = configparser.ConfigParser(interpolation=None)
    config.optionxform = str
    config.read(fname)
    return config


def save_ini(ini_fname, config):
    with open(ini_fname, "w") as file:
        config.write(file)


def load_settings(ini_fname):
    return to_path_settings(laod_ini(ini_fname))


def save_settings(ini_fname, settings):
    save_ini(ini_fname, from_path_settings(settings))



import sys

from ScopeFoundry.data_browser import DataBrowser
from ScopeFoundry.helper_funcs import sibling_path

LOGO_PATH = sibling_path(__file__, "logo.png")

if __name__ == "__main__":
    app = DataBrowser(sys.argv)

    from ScopeFoundry.data_browser.plug_ins.h5_search import H5SearchPlugIn
    from ScopeFoundry.data_browser.plug_ins.time_note import TimeNote

    app.add_plugin(H5SearchPlugIn(app))
    app.add_plugin(TimeNote(app))

    from ScopeFoundry.data_browser.viewers import H5TreeView, RangedOptimizationH5View
    app.add_view(H5TreeView(app))
    app.add_view(RangedOptimizationH5View(app))

    app.set_logo(LOGO_PATH)
    app.settings_load_ini("defaults.ini")

    sys.exit(app.exec_())




'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''
import pyqtgraph as pg
from pyqtgraph.dockarea import DockArea
from qtpy.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget

from ScopeFoundry import Measurement, h5_io


class $READOUT_CLASS_NAME(Measurement):

    name = "$READOUT_NAME"

    def setup(self):

        S = self.settings
        S.New("save_h5", bool, initial=True)

        # just making a guess of data structure here to test if display works
        self.data = {
            "y": [2, 5, 2]
        }

    def setup_figure(self):
        self.ui = DockArea()
        widget = QWidget()
        layout = QVBoxLayout(widget)
        self.plot_dock = self.ui.addDock(
            name=self.name, widget=widget, position='right')

        # setup measurement controlls
        cb_layout = QHBoxLayout()
        layout.addLayout(cb_layout)
        cb_layout.addWidget(self.settings.New_UI(exclude=('activation', 'run_state', 'profile', )))
        cb_layout.addWidget(self.settings.activation.new_pushButton())

        # setup a plot
        graph_layout = pg.GraphicsLayoutWidget(border=(100, 100, 100))
        layout.addWidget(graph_layout)
        self.plot = graph_layout.addPlot(title=self.name)
        self.plot_lines = {}
        self.plot_lines["y"] = self.plot.plot(pen="g")

        # setup hw controls widget
        hw = self.app.hardware["$HW_NAME"]
        self.ui.addDock(name="$HW_NAME",
                        widget=hw.new_control_widgets(),
                        #position='below'
                        )

        self.update_display()

    def update_display(self):
        self.plot_lines["y"].setData(self.data['y'])

    def run(self):
        print(self.__class__, 'run method not implemented')

        # hw = self.app.hardware["$HW_NAME"]
        # self.data = {"y": []}
        # self.data['y'] = hw.dev.read_data()

        if self.settings['save_h5']:
            self.save_h5_data()

    def save_h5_data(self):
        h5_file = h5_io.h5_base_file(app=self.app, measurement=self)
        h5_meas_group = h5_io.h5_create_measurement_group(self, h5_file)
        for k, v in self.data.items():
            h5_meas_group[k] = v
        h5_file.close()



from ScopeFoundry import Measurement, BaseMicroscopeApp
import numpy as np
import pyqtgraph as pg
import time

class Measure1(Measurement):
    
    name = 'measure1'
    
    M = 5

    def setup(self):
        self.settings.New('nested_interrupt', dtype=bool, initial=True)
    
    def run(self):
        
        m2 = self.app.measurements['measure2']
        
        self.current_data = [0]
        self.data = [None,]*self.M
        
        # start M2
        for j in range(self.M):
            if self.interrupt_measurement_called:
                break
            self.set_progress(100*(j+1)/(self.M+1))
            m2.settings['amplitude'] = j+1            
            success = self.start_nested_measure_and_wait(m2, nested_interrupt=self.settings['nested_interrupt'], 
                                               polling_func=self.m2_polling, polling_time=0.5)
            if not success:
                #self.interrupt()
                return
            self.data[j] = m2.data_array
    
    def m2_polling(self):
        m2 = self.app.measurements['measure2']
        self.current_data = m2.data_array.copy()
    

    def setup_figure(self):
        self.ui = self.plot = pg.PlotWidget()

        self.current_plotline = self.plot.plot(pen='r')
        self.plot_lines = []
        for j in range(self.M):
            self.plot_lines.append(self.plot.plot())

    def update_display(self):
        
        self.current_plotline.setData(self.current_data)
        
        for j in range(self.M):
            if self.data[j] is None:
                self.plot_lines[j].setData([0])
            else:
                self.plot_lines[j].setData(self.data[j])

class Measure2(Measurement):
    
    name = 'measure2'
    
    def setup(self):
        
        self.settings.New('amplitude', dtype=float, initial=1.0)
        self.settings.New('run_crash_immediately', dtype=bool, initial=False)
        self.settings.New('run_crash_middle', dtype=bool, initial=False)
        self.settings.New('pre_run_crash', dtype=bool, initial=False)
        self.settings.New('post_run_crash', dtype=bool, initial=False)
        
        
    def setup_figure(self):
        self.ui = self.plot = pg.PlotWidget()
        
        self.plot_line = self.plot.plot()
        
    
    def run(self):
        
        if self.settings['run_crash_immediately']:
            raise IOError("run_crash_immediately")
        
        N = 100
        
        self.data_array = np.zeros(N)
        
        for i in range(N):
            #print(self.name, i, 'of', N)
            if self.interrupt_measurement_called:
                print(self.name, 'interrupted at', i, 'of', N)
                break            
            self.set_progress(100.0*i/N)
            self.data_array[i] = self.settings['amplitude']*np.sin(2*np.pi*i/N)
            time.sleep(0.1)
            if i>50 and self.settings['run_crash_middle']:
                raise IOError("run_crash_middle")

    
    def update_display(self):
        self.plot_line.setData(self.data_array)
    
    def pre_run(self):
        print(self.name, 'pre_run fun!')
        time.sleep(0.5)
        if self.settings['pre_run_crash']:
            raise IOError("pre_run_crash")        
        time.sleep(0.5)
        print(self.name, 'pre_run fun done!')

    def post_run(self):
        print(self.name, 'post_run fun!')
        time.sleep(0.5)
        if self.settings['post_run_crash']:
            raise IOError("post_run_crash")
        time.sleep(0.5)
        print(self.name, 'post_run fun done!')

class NestMeasureTestApp(BaseMicroscopeApp):
    
    name = 'nested_measure_test'
    
    def setup(self):
        
        self.add_measurement(Measure1(self))
        self.add_measurement(Measure2(self))
        
        
if __name__ == '__main__':
    import sys
    app = NestMeasureTestApp(sys.argv)
    
    app.exec_()
    



'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''

# GOAL OF THIS FILE IS TO HANDLE LOW LEVEL COMMUNICATION THROUGH A SERIAL PORT
# SHOULD BE INDEPENDENT OF ANY SCOPEFOUNDRY FUNCTIONALITY.
# RUN THIS FILE TO TEST CONNECTION, TESTING COMMANDS WHILE CHECK main()

import serial
import time

# GUESS OR FROM DEVICE DOCUMENTATION
NEWLINE = "\r"  # typically "\r", "\r\n" or "\n"


class $DEV_CLASS_NAME:

    def __init__(self,
                 port="COM1",  # on windows see device manager
                 debug=False):

        self.port = port
        self.debug = debug

        # TODO: FROM DEVICE DOCUMENTATION
        baudrate = 115200
        bytesize = 8
        parity = 'N'
        stopbits = 1
        xonxoff = False
        rtscts = True

        timeout = 1.0

        self.ser = serial.Serial(
            port=port,
            baudrate=baudrate,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            xonxoff=xonxoff,
            rtscts=rtscts,
            timeout=timeout
        )

    def write(self, cmd: str):
        if self.debug:
            print("write:", repr(cmd))
        self.ser.write((cmd + NEWLINE).encode())

    def query(self, cmd: str):
        self.write(cmd)
        time.sleep(0.01)
        resp: bytes = self.ser.readline()
        if self.debug:
            print("resp:", resp.decode())
        return resp.decode()

    def close(self):
        self.ser.close()

    def write_property_x(self, value):
        raise NotImplementedError
        # TYPICALLY:
        # self.write(f'SET_PROPERTY_X_CMD {value}')

    def read_property_x(self):
        raise NotImplementedError
        # TYPICALLY:
        # resp = self.query(f'GET_PROPERTY_X_CMD {value}')
        # return 'convert_to_float(resp)'

    def read_data(self):
        # NOTE read_data is a too generic function name
        raise NotImplementedError

        # TYPICALLY:
        # resp = self.query(f'GET_DATA_COMMAND {value}')
        # return a list of data points from resp.


if __name__ == '__main__':
    print('start')
    dev = $DEV_CLASS_NAME(port="COM1",  # TODO: on windows see device manager
                          debug=True)
    print(dev.query('TEST_DOCUMENTED_CMD_HERE'))

    # print(dev.read_property_x())
    # print(dev.read_data())

    print('done')



import unittest
from ScopeFoundry.measurement import Measurement
from ScopeFoundry import BaseMicroscopeApp


class DummyMeasure(Measurement):
    name = "dummy"

    def setup(self):
        self.settings.New("choices", int, choices=INITIAL_CHOICES)


INITIAL_CHOICES = [("choice 1", 1), ("choice 2", 2)]


class ChoiceListTest(unittest.TestCase):
    
    def setUp(self):
        self.app = BaseMicroscopeApp([])
        self.app.add_measurement(DummyMeasure(self))
        
    def test_initial_value_set(self): 
        self.assertEqual(self.app.measurements.dummy.settings["choices"], 1)        

    def test_change_choice_list_all_new_choices(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.change_choice_list([("choice 3", 3)])  
        self.assertEqual(lq.val, 3)

    def test_change_choice_list_containing_oldval(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.change_choice_list([("choice 1 renamed", 1), ("choice 3", 3)])  
        self.assertEqual(lq.val, 1)
    
    def test_change_choice_list_all_new(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.change_choice_list([("choice 3", 3), ("choice 4", 4)])  
        self.assertEqual(lq.val, 3)    
        
    def test_change_choice_list_empty(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.change_choice_list([])  
        self.assertEqual(lq.val, 1)  # for now expected behavior

    def test_add_choices(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.add_choices([("choice 4", 4)])
        self.assertEqual(lq.val, 1)
        self.assertIn(("choice 4", 4), lq.choices)

    def test_add_choices_overlapping(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.add_choices([("choice 2", 2), ("choice 4", 4)])
        self.assertEqual(lq.val, 1)
        self.assertIn(("choice 1", 1), lq.choices)
        self.assertIn(("choice 2", 2), lq.choices)
        self.assertIn(("choice 4", 4), lq.choices)

    def test_add_choices_and_set(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.add_choices([("choice 4", 4)], new_val=2)
        self.assertEqual(lq.val, 2)

    def test_add_empty(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.add_choices([])
        self.assertEqual(lq.val, 1)  # for now expected behavior

    def test_remove_a_set_choice(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.remove_choices([("choice 1", 1)])
        self.assertEqual(lq.val, 2) 
        
    def test_remove_unset_choice(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.remove_choices([("choice 2", 2)])
        self.assertEqual(lq.val, 1)
   
    def test_remove_all(self): 
        lq = self.app.measurements.dummy.settings.get_lq("choices")
        lq.remove_choices(INITIAL_CHOICES)
        self.assertEqual(lq.val, 1)  # for now expected behavior

                                
if __name__ == '__main__':
    unittest.main()
    



'''
Created on $DATE_PRETTY

@author: $AUTHORS
'''
from ScopeFoundry.hardware import HardwareComponent


class $HW_CLASS_NAME(HardwareComponent):

    name = "$HW_NAME"



    def setup(self):
        S = self.settings
        S.New("GPIB_address", str, initial="GPIB0::27::INSTR")
        S.New("property_x", float, ro=False, unit='mm', spinbox_decimals=4, si=False)
                          

    def connect(self):
        if hasattr(self, 'dev'):
            return

        S = self.settings

        import pyvisa
        rm = pyvisa.ResourceManager()
        self.dev = rm.open_resource(S["GPIB_address"])
        # attache read and write functions to settings
        # S.property_x.connect_to_hardware(
        #     self.read_property_x,
        #     self.write_property_x)

    def disconnect(self):
        if not hasattr(self, 'dev'):
            return

        # self.settings.disconnect_all_from_hardware()
        self.dev.close()
        del self.dev


    def write_property_x(self, value):
        raise NotImplementedError
        # TYPICALLY:
        # return self.dev.query('SET_PROPERTY_X_CMD_FROM_DOCUMENTATION')

    def read_property_x(self):
        raise NotImplementedError
        # TYPICALLY:
        # return self.dev.query('GET_PROPERTY_X_CMD_FROM_DOCUMENTATION')


    def read_data(self):
        # NOTE read_data is a too generic function name
        raise NotImplementedError

        # TYPICALLY:
        # return self.dev.query('GET_PROPERTY_X_CMD_FROM_DOCUMENTATION')


    # if you want to continuously update settings implement *run* method
    # def run(self):
    #     self.settings.property_x.read_from_hardware()




import datetime
from pathlib import Path
from qtpy import QtCore, QtWidgets

from ScopeFoundry.data_browser.data_browser_plug_in import DataBrowserPlugIn


class TimeNote(DataBrowserPlugIn):
    name = "time_note"
    button_text = "🕑Note"
    show_keyboard_key = QtCore.Qt.Key_N
    description = "inject simple notes with time stamps (Ctrl + N)"

    def setup(self):
        self.subject_line = QtWidgets.QLineEdit()

        subj_layout = QtWidgets.QHBoxLayout()
        subj_layout.addWidget(QtWidgets.QLabel("🕑 subject"))
        subj_layout.addWidget(self.subject_line)

        self.text_edit = QtWidgets.QTextEdit()
        self.subject_line.setText("")
        self.create_before_btn = QtWidgets.QPushButton("create note before selected")
        self.create_btn = QtWidgets.QPushButton("create note now")

        btn_layout = QtWidgets.QHBoxLayout()
        btn_layout.addWidget(self.create_before_btn)
        btn_layout.addWidget(self.create_btn)

        self.ui = QtWidgets.QWidget(objectName="NoteWidget")
        self.ui.setMaximumHeight(1200)
        layout = QtWidgets.QVBoxLayout(self.ui)
        layout.addLayout(subj_layout)
        layout.addWidget(self.text_edit)
        layout.addLayout(btn_layout)
        self.ui.setStyleSheet(
            "QWidget#NoteWidget{background-color:rgba(216, 120, 152, 0.1)}"
        )

        self.create_before_btn.clicked.connect(self.new_before_selected_file)
        self.create_btn.clicked.connect(self.new_now)

    def update(self, fname: str=None) -> None:
        if not fname.endswith(".txt"):
            return

        subject = parse_subject(fname)
        if subject:
            self.subject_line.setText(subject)
        with open(fname, "r", encoding="utf8") as f:
            self.text_edit.setText(f.read())

    def new_now(self):
        time_str = "{timestamp:%y%m%d_%H%M%S}".format(timestamp=datetime.datetime.now())
        self.new_note(time_str)

    def new_before_selected_file(self):
        fname = self.new_fname
        time_str = parse_timestampe_before(fname)

        if time_str is False:
            print(
                fname, "not supported, select a file with name pattern YYMMDD_HHMMSS_*"
            )
            return

        self.new_note(time_str)

    def new_note(self, time_str):
        subject = self.subject_line.text()
        new_fname = (
            Path(self.new_fname).parent.absolute()
            / f"{time_str}_{subject.replace(' ', '_')}.txt"
        )

        with open(new_fname, "w", encoding="utf8") as f:
            f.write(self.text_edit.toPlainText())

        print("made time note", new_fname)


def parse_timestampe_before(fname):
    parts = Path(fname).stem.split("_")

    if len(parts) >= 2 and parts[0].isnumeric() and parts[1].isnumeric():
        return f"{parts[0]}_{int(parts[1])-1}"

    return False


def parse_subject(fname):
    parts = Path(fname).stem.split("_")

    if len(parts) >= 2 and parts[0].isnumeric() and parts[1].isnumeric():
        return "_".join(parts[2:])

    return False



from .import_from_gh_qt_view import ImportFromGhView as View
from .utils import InfoTypes, decorate_info_text, run_subprocess


class ImportFromGhController:
    def __init__(self, view: View):
        self.view = view
        # self.view.validate_my_gh_acc_validate_clicked(self.validate_my_gh_acc)
        self.view.validate_origin_repo_clicked(self.validate_origin_repo)
        self.view.fork_clicked(self.fork_on_gh)
        self.view.import_clicked(self.import_from_origin)
        self.view.validate_fork_clicked(self.validate_fork)
        self.view.import_from_fork_clicked(self.import_from_fork)

        self.view.set_ready_to_import(False)
        self.view.set_ready_to_import_from_fork(False)

    def set_msg(self, text, info_type=InfoTypes.INFO):
        text = decorate_info_text(text, info_type)
        self.view.set_info_text(text)

    def append_msg(self, text, info_type=InfoTypes.INFO):
        text = decorate_info_text(text, info_type)
        self.view.append_info(text)

    def new_task_msg(self, text):
        self.set_msg(text, InfoTypes.TASK)

    def validate_my_gh_acc(self):
        raise NotImplementedError
        # no necessary
        # self.set_info('validating my acc')
        # self.view.set_my_gh_acc_valid(True)

    def validate_origin_repo(self):
        origin_repo = self.view.get_origin_repo()
        self.new_task_msg(f'validating origin repo {origin_repo}')
        success = self.check_exists_and_report(origin_repo)
        # self.view.set_origin_repo_valid(success)
        self.view.set_ready_to_import(success)

    def import_from_origin(self):
        origin_repo = self.view.get_origin_repo()
        name = origin_repo.split('/')[-1].strip('.git').strip('HW_')
        destination_dir = f"ScopeFoundryHW/{name}"
        self.import_repo(origin_repo, destination_dir)

    def fork_on_gh(self):
        origin_repo = self.view.get_origin_repo()

        stdout, stderr = fork_on_gh(origin_repo)
        self.append_msg(stdout, InfoTypes.INFO)
        self.append_msg(stderr, InfoTypes.ERROR)
        self.validate_fork()

    def import_from_fork(self):
        origin_repo = self.view.get_origin_repo()
        name = origin_repo.split('/')[-1].strip('.git').strip('HW_')
        my_gh_acc = self.view.get_my_gh_acc()

        repo_to_import = f"https://github.com/{my_gh_acc}/HW_{name}.git"
        destination_dir = f"ScopeFoundryHW/{name}"
        self.new_task_msg(f'importing {repo_to_import} to {destination_dir}')
        self.import_repo(repo_to_import, destination_dir)

    def validate_fork(self):
        my_gh_acc = self.view.get_my_gh_acc()
        origin_repo = self.view.get_origin_repo()
        module_name = origin_repo.split('/')[-1].strip('.git')

        forked_repo = f"https://github.com/{my_gh_acc}/{module_name}"
        self.new_task_msg(f'validating fork repo {forked_repo}')
        success = self.check_exists_and_report(forked_repo)
        self.view.set_ready_to_import_from_fork(success)

    def import_repo(self, repo_to_import, destination_dir):
        remote_name = repo_to_import.split('/')[-1].strip('.git')
        stdout, stderr = subtree_merge_strategy(
            remote_name, repo_to_import, destination_dir)
        self.append_msg(stdout, InfoTypes.INFO)
        self.append_msg(stderr, InfoTypes.ERROR)

    def check_exists_and_report(self, repo):
        stdout, stderr = ls_repo(repo)
        exists = stderr == ""
        if exists:
            self.append_msg(f"repo exists: {repo}", InfoTypes.SUCCESS)
        else:
            self.append_msg(f"repo does NOT exist {repo}", InfoTypes.FAILED)
            # self.append_info(stdout, InfoTypes.INFO)
            self.append_msg(stderr, InfoTypes.ERROR)
        return exists


def fork_on_gh(origin_repo):
    return run_subprocess(f"gh repo fork {origin_repo} --remote=True")


def ls_repo(repo="https://github.com/ScopeFoundry/HW_picam.git"):
    cmd = f"git ls-remote {repo}"
    return run_subprocess(cmd)


def subtree_merge_strategy(remote_name, repo_to_import, destination_dir):
    cmd = f"bash scripts/merge_remote_tree_with_subdir.sh {remote_name} {repo_to_import} {destination_dir}"
    return run_subprocess(cmd)



from qtpy import QtWidgets
from ScopeFoundry.data_browser import DataBrowserView
import os

class NPZView(DataBrowserView):
    """Reads Numpy Z files (npz)
    
    """
    name = 'npz_view'
    
    def setup(self):
        
        #self.ui = QtGui.QScrollArea()
        #self.display_label = QtGui.QLabel("TestNPZView")
        self.ui = self.display_textEdit = QtWidgets.QTextEdit()
        
        #self.ui.setLayout(QtGui.QVBoxLayout())
        #self.ui.layout().addWidget(self.display_label)
        #self.ui.setWidget(self.display_label)
        
    def on_change_data_filename(self, fname=None):
        import numpy as np
        
        try:
            self.dat = np.load(fname)
            
            self.display_txt = "File: {}\n".format(fname)
            
            sorted_keys = sorted(self.dat.keys())
            
            for key in sorted_keys:
                val = self.dat[key]
                if val.shape == ():
                    self.display_txt += "    --> {}: {}\n".format(key, val)                    
                else:
                    self.display_txt += "    --D {}: Array of {} {}\n".format(key, val.dtype, val.shape)
            
            #self.display_label.setText(self.display_txt)
            self.display_textEdit.setText(self.display_txt)
        except Exception as err:
            self.display_textEdit.setText("failed to load %s:\n%s" %(fname, err))
            raise(err)
        
    def is_file_supported(self, fname):
        return os.path.splitext(fname)[1] == ".npz"




import subprocess

import argparse


def main():
    
    parser = argparse.ArgumentParser(description='Grab a Hardware Component from GitHub and place it in ScopeFoundryHW using git subtree add/pull.')

    parser.add_argument('name', type=str,
                    help='name of hardware')
    
    parser.add_argument('action', help='add or pull')
        
    parser.add_argument('--squash', action='store_true', 
                            required=False, help='')
    

    args = parser.parse_args()
    print(args)
    print("-->", args['name'])
    
    
if __name__ == "__main__":
    main()
